import itertools
import logging
from typing import Dict, Union

import casadi as ca

import numpy as np

import pkg_resources

import pymoca
import pymoca.backends.casadi.api

from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools._internal.casadi_helpers import substitute_in_external

from .optimization_problem import OptimizationProblem
from .timeseries import Timeseries

logger = logging.getLogger("rtctools")


class ModelicaMixin(OptimizationProblem):
    """
    Adds a `Modelica <http://www.modelica.org/>`_ model to your optimization problem.

    During preprocessing, the Modelica files located inside the ``model`` subfolder are loaded.

    :cvar modelica_library_folders:
        Folders in which any referenced Modelica libraries are to be found. Default is an empty list.
    """

    # Folders in which the referenced Modelica libraries are found
    modelica_library_folders = []

    def __init__(self, **kwargs):
        # Check arguments
        assert('model_folder' in kwargs)

        # Log pymoca version
        logger.debug("Using pymoca {}.".format(pymoca.__version__))

        # Transfer model from the Modelica .mo file to CasADi using pymoca
        if 'model_name' in kwargs:
            model_name = kwargs['model_name']
        else:
            if hasattr(self, 'model_name'):
                model_name = self.model_name
            else:
                model_name = self.__class__.__name__

        self.__pymoca_model = pymoca.backends.casadi.api.transfer_model(
            kwargs['model_folder'], model_name, self.compiler_options())

        # Extract the CasADi MX variables used in the model
        self.__mx = {}
        self.__mx['time'] = [self.__pymoca_model.time]
        self.__mx['states'] = [v.symbol for v in self.__pymoca_model.states]
        self.__mx['derivatives'] = [v.symbol for v in self.__pymoca_model.der_states]
        self.__mx['algebraics'] = [v.symbol for v in self.__pymoca_model.alg_states]
        self.__mx['parameters'] = [v.symbol for v in self.__pymoca_model.parameters]
        self.__mx['string_parameters'] = [v.name for v in (*self.__pymoca_model.string_parameters,
                                                           *self.__pymoca_model.string_constants)]
        self.__mx['control_inputs'] = []
        self.__mx['constant_inputs'] = []
        self.__mx['lookup_tables'] = []

        # Merge with user-specified delayed feedback
        for v in self.__pymoca_model.inputs:
            if v.symbol.name() in self.__pymoca_model.delay_states:
                # Delayed feedback variables are local to each ensemble, and
                # therefore belong to the collection of algebraic variables,
                # rather than to the control inputs.
                self.__mx['algebraics'].append(v.symbol)
            else:
                if v.symbol.name() in kwargs.get('lookup_tables', []):
                    self.__mx['lookup_tables'].append(v.symbol)
                elif v.fixed:
                    self.__mx['constant_inputs'].append(v.symbol)
                else:
                    self.__mx['control_inputs'].append(v.symbol)

        # Initialize nominals and types
        # These are not in @cached dictionary properties for backwards compatibility.
        self.__python_types = AliasDict(self.alias_relation)
        for v in itertools.chain(
                    self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs):
            self.__python_types[v.symbol.name()] = v.python_type

        # Initialize dae, initial residuals, as well as delay arguments
        # These are not in @cached dictionary properties so that we need to create the list
        # of function arguments only once.
        variable_lists = ['states', 'der_states', 'alg_states', 'inputs', 'constants', 'parameters']
        function_arguments = [self.__pymoca_model.time] + [
            ca.veccat(*[v.symbol for v in getattr(self.__pymoca_model, variable_list)])
            for variable_list in variable_lists]

        self.__dae_residual = self.__pymoca_model.dae_residual_function(*function_arguments)
        if self.__dae_residual is None:
            self.__dae_residual = ca.MX()

        self.__initial_residual = self.__pymoca_model.initial_residual_function(*function_arguments)
        if self.__initial_residual is None:
            self.__initial_residual = ca.MX()

        # Log variables in debug mode
        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("ModelicaMixin: Found states {}".format(
                ', '.join([var.name() for var in self.__mx['states']])))
            logger.debug("ModelicaMixin: Found derivatives {}".format(
                ', '.join([var.name() for var in self.__mx['derivatives']])))
            logger.debug("ModelicaMixin: Found algebraics {}".format(
                ', '.join([var.name() for var in self.__mx['algebraics']])))
            logger.debug("ModelicaMixin: Found control inputs {}".format(
                ', '.join([var.name() for var in self.__mx['control_inputs']])))
            logger.debug("ModelicaMixin: Found constant inputs {}".format(
                ', '.join([var.name() for var in self.__mx['constant_inputs']])))
            logger.debug("ModelicaMixin: Found parameters {}".format(
                ', '.join([var.name() for var in self.__mx['parameters']])))

        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

    @cached
    def compiler_options(self) -> Dict[str, Union[str, bool]]:
        """
        Subclasses can configure the `pymoca <http://github.com/pymoca/pymoca>`_ compiler options here.

        :returns: A dictionary of pymoca compiler options.  See the pymoca documentation for details.
        """

        # Default options
        compiler_options = {}

        # Expand vector states to multiple scalar component states.
        compiler_options['expand_vectors'] = True

        # Where imported model libraries are located.
        library_folders = self.modelica_library_folders.copy()

        for ep in pkg_resources.iter_entry_points(group='rtctools.libraries.modelica'):
            if ep.name == "library_folder":
                library_folders.append(
                    pkg_resources.resource_filename(ep.module_name, ep.attrs[0]))

        compiler_options['library_folders'] = library_folders

        # Eliminate equations of the type 'var = const'.
        compiler_options['eliminate_constant_assignments'] = True

        # Eliminate constant symbols from model, replacing them with the values
        # specified in the model.
        compiler_options['replace_constant_values'] = True

        # Replace any constant expressions into the model.
        compiler_options['replace_constant_expressions'] = True

        # Replace any parameter expressions into the model.
        compiler_options['replace_parameter_expressions'] = True

        # Eliminate variables starting with underscores.
        compiler_options['eliminable_variable_expression'] = r'(.*[.]|^)_\w+(\[[\d,]+\])?\Z'

        # Pymoca currently requires `expand_mx` to be set for
        # `eliminable_variable_expression` to work.
        compiler_options['expand_mx'] = True

        # Automatically detect and eliminate alias variables.
        compiler_options['detect_aliases'] = True

        # Disallow aliasing to derivative states
        compiler_options['allow_derivative_aliases'] = False

        # Cache the model on disk
        compiler_options['cache'] = True

        # Done
        return compiler_options

    def delayed_feedback(self):
        delayed_feedback = super().delayed_feedback()

        # Create delayed feedback
        for delay_state, delay_argument in zip(self.__pymoca_model.delay_states, self.__pymoca_model.delay_arguments):
            delayed_feedback.append(
                (delay_argument.expr, delay_state, delay_argument.duration))
        return delayed_feedback

    @property
    def dae_residual(self):
        return self.__dae_residual

    @property
    def dae_variables(self):
        return self.__mx

    @property
    @cached
    def output_variables(self):
        output_variables = [ca.MX.sym(variable)
                            for variable in self.__pymoca_model.outputs]
        output_variables.extend(self.__mx['control_inputs'])
        return output_variables

    @cached
    def parameters(self, ensemble_member):
        # Call parent class first for default values.
        parameters = super().parameters(ensemble_member)

        # Return parameter values from pymoca model
        parameters.update({v.symbol.name(): v.value for v in self.__pymoca_model.parameters})

        # Done
        return parameters

    @cached
    def string_parameters(self, ensemble_member):
        # Call parent class first for default values.
        parameters = super().string_parameters(ensemble_member)

        # Return parameter values from pymoca model
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_parameters})
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_constants})

        # Done
        return parameters

    @cached
    def history(self, ensemble_member):
        history = super().history(ensemble_member)

        initial_time = np.array([self.initial_time])

        # Parameter values
        parameters = self.parameters(ensemble_member)
        parameter_values = [parameters.get(param.name(), param) for param in self.__mx['parameters']]

        # Initial conditions obtained from start attributes.
        for v in self.__pymoca_model.states:
            if v.fixed:
                sym_name = v.symbol.name()
                start = v.start

                if isinstance(start, ca.MX):
                    # If start contains symbolics, try substituting parameter values
                    if isinstance(start, ca.MX) and not start.is_constant():
                        [start] = substitute_in_external([start], self.__mx['parameters'], parameter_values)
                        if not start.is_constant() or np.isnan(float(start)):
                            raise Exception('ModelicaMixin: Could not resolve initial value for {}'.format(sym_name))

                    start = v.python_type(start)

                history[sym_name] = Timeseries(initial_time, start)

                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("ModelicaMixin: Initial state variable {} = {}".format(sym_name, start))

        return history

    @property
    def initial_residual(self):
        return self.__initial_residual

    @cached
    def bounds(self):
        # Call parent class first for default values.
        bounds = super().bounds()

        # Parameter values
        parameters = self.parameters(0)
        parameter_values = [parameters.get(param.name(), param) for param in self.__mx['parameters']]

        # Load additional bounds from model
        for v in itertools.chain(
                self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs):
            sym_name = v.symbol.name()

            try:
                (m, M) = bounds[sym_name]
            except KeyError:
                if self.__python_types.get(sym_name, float) == bool:
                    (m, M) = (0, 1)
                else:
                    (m, M) = (-np.inf, np.inf)

            m_ = v.min
            if isinstance(m_, ca.MX) and not m_.is_constant():
                [m_] = substitute_in_external([m_], self.__mx['parameters'], parameter_values)
                if not m_.is_constant() or np.isnan(float(m_)):
                    raise Exception('Could not resolve lower bound for variable {}'.format(sym_name))
            m_ = float(m_)

            M_ = v.max
            if isinstance(M_, ca.MX) and not M_.is_constant():
                [M_] = substitute_in_external([M_], self.__mx['parameters'], parameter_values)
                if not M_.is_constant() or np.isnan(float(M_)):
                    raise Exception('Could not resolve upper bound for variable {}'.format(sym_name))
            M_ = float(M_)

            # We take the intersection of all provided bounds
            m = max(m, m_)
            M = min(M, M_)

            bounds[sym_name] = (m, M)

        return bounds

    @cached
    def seed(self, ensemble_member):
        # Call parent class first for default values.
        seed = super().seed(ensemble_member)

        # Parameter values
        parameters = self.parameters(ensemble_member)
        parameter_values = [parameters.get(param.name(), param) for param in self.__mx['parameters']]

        # Load seeds
        for var in itertools.chain(self.__pymoca_model.states, self.__pymoca_model.alg_states):
            if var.fixed:
                # Values will be set from import timeseries
                continue

            start = var.start

            if isinstance(start, ca.MX) or start != 0.0:
                sym_name = var.symbol.name()

                # If start contains symbolics, try substituting parameter values
                if isinstance(start, ca.MX) and not start.is_constant():
                    [start] = substitute_in_external([start], self.__mx['parameters'], parameter_values)
                    if not start.is_constant() or np.isnan(float(start)):
                        logger.error('ModelicaMixin: Could not resolve seed value for {}'.format(sym_name))
                        continue

                times = self.times(sym_name)
                start = var.python_type(start)
                s = Timeseries(times, np.full_like(times, start))
                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("ModelicaMixin: Seeded variable {} = {}".format(sym_name, start))
                seed[sym_name] = s

        return seed

    def variable_is_discrete(self, variable):
        return self.__python_types.get(variable, float) != float

    @property
    @cached
    def alias_relation(self):
        return self.__pymoca_model.alias_relation

    @property
    @cached
    def __nominals(self):

        # Make the dict
        nominal_dict = AliasDict(self.alias_relation)

        # Grab parameters and their values
        parameters = self.parameters(0)
        parameter_values = [parameters.get(param.name(), param) for param in self.__mx['parameters']]

        # Iterate over nominalizable states
        for v in itertools.chain(
                self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs):
            sym_name = v.symbol.name()
            nominal = v.nominal

            # If nominal contains parameter symbols, substitute them
            if isinstance(nominal, ca.MX) and not nominal.is_constant():
                [nominal] = substitute_in_external([nominal], self.__mx['parameters'], parameter_values)
                if not nominal.is_constant() or np.isnan(float(nominal)):
                    logger.error('ModelicaMixin: Could not resolve nominal value for {}'.format(sym_name))
                    continue

            nominal = float(nominal)

            if not np.isnan(nominal):
                # Take absolute value (nominal sign is meaningless- a nominal is a magnitude)
                nominal = abs(nominal)

                # If nominal is 0 or 1, we just use the default (1.0)
                if nominal in (0.0, 1.0):
                    continue

                nominal_dict[sym_name] = nominal

                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("ModelicaMixin: Set nominal value for variable {} to {}".format(
                        sym_name, nominal))
            else:
                logger.warning("ModelicaMixin: Could not set nominal value for {}".format(sym_name))

        return nominal_dict

    def variable_nominal(self, variable):
        try:
            return self.__nominals[variable]
        except KeyError:
            return super().variable_nominal(variable)



def cached(f):
    def wrapper(self, ensemble_member=None):
        def call():
            if ensemble_member is not None:
                return f(self, ensemble_member)
            else:
                return f(self)
        # Add a check so that caching is applied to the 'toplevel'
        # method implementation in the class hierarchy only.
        call_in_progress = '__' + f.__name__ + '_in_progress'
        if hasattr(self, call_in_progress):
            return call()
        cache_name = '__' + f.__name__
        if ensemble_member is not None:
            cache_name = '{}[{}]'.format(cache_name, ensemble_member)
        if hasattr(self, cache_name):
            return getattr(self, cache_name)
        setattr(self, call_in_progress, True)
        value = call()
        setattr(self, cache_name, value)
        delattr(self, call_in_progress)
        return value
    return wrapper



import os
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Iterable, List, Union

try:
    from netCDF4 import Dataset, Variable, chartostring
except ImportError:
    raise ImportError("NetCDF4 is required when using NetCDF/NetCDFMixin")

try:
    from cftime import num2pydate as num2date
except ImportError:
    from cftime import num2date

import numpy as np


class Stations:

    def __init__(self, dataset: Dataset, station_variable: Variable):
        self.__station_variable = station_variable

        station_dimension = station_variable.dimensions[0]

        # todo make this a bit smarter, right now variables like station_name would be forgotten
        self.__attribute_variables = {}
        for variable_name in dataset.variables:
            variable = dataset.variables[variable_name]
            if variable != station_variable and variable.dimensions == (station_dimension,):
                self.__attribute_variables[variable_name] = variable

        self.__attributes = OrderedDict()
        for i in range(station_variable.shape[0]):
            id = str(chartostring(station_variable[i]))

            values = {}
            for variable_name in self.__attribute_variables.keys():
                values[variable_name] = dataset.variables[variable_name][i]

            self.__attributes[id] = values

    @property
    def station_ids(self) -> Iterable:
        """
        :return: An ordered iterable of the station ids (location ids) for which station data is available.
        """
        return self.__attributes.keys()

    @property
    def attributes(self) -> OrderedDict:
        """
        :return: An OrderedDict containing dicts containing the values for all station attributes of the input dataset.
        """
        return self.__attributes

    @property
    def attribute_variables(self) -> dict:
        """
        :return: A dict containing the station attribute variables of the input dataset.
        """
        return self.__attribute_variables


class ImportDataset:
    """
    A class used to open and import the data from a NetCDF file.
    Uses the NetCDF4 library. Contains various methods for reading the data in the file.
    """

    def __init__(self, folder: str, basename: str):
        """
        :param folder:    Folder the file is located in.
        :param basename:  Basename of the file, extension ".nc" will be appended to this
        """

        self.__ensemble_size = 1

        # Load the content of a NetCDF file into a Dataset.
        self.__filename = os.path.join(
            folder,
            basename + ".nc"
        )
        self.__dataset = Dataset(self.__filename)

        # Find the number of ensemble members and the time and station id variables
        self.__time_variable = self.__find_time_variable()
        if self.__time_variable is None:
            raise Exception('No time variable found in file ' + self.__filename + '. '
                            'Please ensure the file contains a time variable with standard_name "time" and axis "T".')

        self.__ensemble_member_variable = self.__find_ensemble_member_variable()
        if self.__ensemble_member_variable:
            self.__ensemble_size = self.__dataset.dimensions['realization'].size

        self.__station_variable = self.__find_station_variable()
        if self.__station_variable is None:
            raise Exception('No station variable found in file ' + self.__filename + '. '
                            'Please ensure the file contains a variable with cf_role "timeseries_id".')

    def __str__(self):
        return self.__filename

    def __find_time_variable(self) -> Union[Variable, None]:
        """
        Find the variable containing the times in the given Dataset.

        :param dataset: The Dataset to be searched.
        :return: a netCDF4.Variable object of the time variable (or None if none found)
        """
        for variable in self.__dataset.variables.values():
            if ('standard_name' in variable.ncattrs() and 'axis' in variable.ncattrs()
                    and variable.standard_name == 'time' and variable.axis == 'T'):
                return variable

        return None

    def __find_ensemble_member_variable(self) -> Union[Variable, None]:
        """
                Find the variable containing the ensemble member index in the given Dataset.

                :param dataset: The Dataset to be searched.
                :return: a netCDF4.Variable object of the ensemble member index variable (or None if none found)
        """
        for variable in self.__dataset.variables.values():
            if 'standard_name' in variable.ncattrs() and variable.standard_name == "realization":
                return variable

        return None

    def __find_station_variable(self) -> Union[Variable, None]:
        """
        Find the variable containing station id's  (location id's) in the given Dataset.

        :param dataset: The Dataset to be searched.
        :return: a netCDF4.Variable object of the station id variable (or None if none found)
        """
        for variable in self.__dataset.variables.values():
            if 'cf_role' in variable.ncattrs() and variable.cf_role == 'timeseries_id':
                return variable

        return None

    def read_import_times(self) -> np.ndarray:
        """
        Reads the import times in the time variable of the dataset.

        :param time_variable: The time variable containing input times
        :return: an array containing the input times as datetime objects
        """
        time_values = self.__time_variable[:]
        time_unit = self.__time_variable.units
        try:
            time_calendar = self.__time_variable.calendar
        except AttributeError:
            time_calendar = u'gregorian'

        return num2date(time_values, units=time_unit, calendar=time_calendar)

    def read_station_data(self) -> Stations:
        return Stations(self.__dataset, self.__station_variable)

    def find_timeseries_variables(self) -> List[str]:
        """
        Find the keys of all 2-D or 3-D variables with dimensions {station, time} or {station, time, realization}
        where station is the dimension of the station_variable, time the dimension of the time_variable and
        realization the dimension for ensemble_member_index.

        :param dataset:           The Dataset to be searched.
        :param station_variable:  The station id variable.
        :param time_variable:     The time variable.
        :return: a list of strings containing all keys found.
        """
        station_dim = self.__station_variable.dimensions[0]
        time_dim = self.__time_variable.dimensions[0]
        if self.__ensemble_member_variable is not None:
            ensemble_dim = self.__ensemble_member_variable.dimensions[0]
            expected_dims = [
                (time_dim, station_dim, ensemble_dim), (time_dim, ensemble_dim, station_dim),
                (station_dim, time_dim, ensemble_dim), (station_dim, ensemble_dim, time_dim),
                (ensemble_dim, time_dim, station_dim), (ensemble_dim, station_dim, time_dim)] \
                + [(station_dim, time_dim), (time_dim, station_dim)]
        else:
            expected_dims = [(station_dim, time_dim), (time_dim, station_dim)]

        timeseries_variables = []
        for var_key, variable in self.__dataset.variables.items():
            if variable.dimensions in expected_dims:
                timeseries_variables.append(var_key)

        return timeseries_variables

    def read_timeseries_values(self, station_index: int, variable_name: str, ensemble_member: int = 0) -> np.ndarray:
        """
        Reads the specified timeseries from the input file.

        :param station_index: The index of the station for which the values should be read
        :param variable_name: The name of the variable for which the values should be read
        :return: an array of values
        """

        station_dim = self.__station_variable.dimensions[0]
        timeseries_variable = self.__dataset.variables[variable_name]

        # possibly usefull for in a debugger mode
        # assert set(timeseries_variable.dimensions)==set(('time', 'station')) \
        #        or set(timeseries_variable.dimensions)==set(('time', 'station', 'realization'))

        if self.__ensemble_member_variable is not None and 'realization' in timeseries_variable.dimensions:
            ensemble_member_dim = self.__ensemble_member_variable.dimensions[0]
            for i in range(3):
                if timeseries_variable.dimensions[i] == station_dim:
                    station_arg_Index = i
                elif timeseries_variable.dimensions[i] == ensemble_member_dim:
                    ensemble_arg_Index = i
            time_arg_Index = set(range(3)) - {station_arg_Index, ensemble_arg_Index}
            time_arg_Index = time_arg_Index.pop()
            argument = [None]*3
            argument[station_arg_Index] = station_index
            argument[ensemble_arg_Index] = ensemble_member
            argument[time_arg_Index] = slice(None)
            values = timeseries_variable[tuple(argument)]
        else:
            if timeseries_variable.dimensions[0] == station_dim:
                values = timeseries_variable[station_index, :]
            else:
                values = timeseries_variable[:, station_index]

        # NetCDF4 reads the values as a numpy masked array,
        # convert to a normal array with nan where mask == True
        return np.ma.filled(values, np.nan)

    def variable_dimensions(self, variable):
        return self.__dataset.variables[variable].dimensions

    @property
    def time_variable(self):
        return self.__time_variable

    @property
    def station_variable(self):
        return self.__station_variable

    @property
    def ensemble_member_variable(self):
        return self.__ensemble_member_variable

    @property
    def ensemble_size(self):
        """
        Ensemble size.
        """
        return self.__ensemble_size


class ExportDataset:
    """
    A class used to write data to a NetCDF file.
    Creates a new file or overwrites an old file. The file metadata will be written upon initialization. Data such
    as times, station data and timeseries data should be presented to the ExportDataset through the various methods.
    When all data has been written, the close method must be called to flush the changes from local memory to the
    actual file on disk.
    """

    def __init__(self, folder: str, basename: str):
        """
        :param folder:   Folder the file will be located in.
        :param basename: Basename of the file, extension ".nc" will be appended to this
        """
        # Create the file and open a Dataset to access it
        self.__filename = os.path.join(
            folder,
            basename + ".nc"
        )
        # use same write format as FEWS
        self.__dataset = Dataset(self.__filename, mode='w', format='NETCDF3_CLASSIC')

        # write metadata to the file
        self.__dataset.title = 'RTC-Tools Output Data'
        self.__dataset.institution = 'Deltares'
        self.__dataset.source = 'RTC-Tools'
        self.__dataset.history = 'Generated on {}'.format(datetime.now())
        self.__dataset.Conventions = 'CF-1.6'
        self.__dataset.featureType = 'timeseries'

        # dimensions are created when writing times and station data, must be created before writing variables
        self.__time_dim = None
        self.__station_dim = None
        self.__station_id_to_index_mapping = None
        self.__ensemble_member_index_dim = None

        self.__timeseries_variables = {}

    def __str__(self):
        return self.__filename

    def write_times(self, times: np.ndarray, forecast_time: float, forecast_date: datetime) -> None:
        """
        Writes a time variable to the given dataset.

        :param dataset:        The NetCDF4.Dataset object that the times will be written to (must have write permission)
        :param times:          The times that are to be written in seconds.
        :param forecast_time:  The forecast time in seconds corresponding to the forecast date
        :param forecast_date:  The datetime corresponding with time in seconds at the forecast index.
        """

        # in a NetCDF file times are written with respect to a reference date
        # the written values for the times may never be negative, so use the earliest time as the reference date
        reference_date = forecast_date
        minimum_time = np.min(times)
        if minimum_time < 0:
            times = times - minimum_time
            reference_date = reference_date - timedelta(seconds=forecast_time - minimum_time)

        self.__time_dim = self.__dataset.createDimension('time', None)

        time_var = self.__dataset.createVariable('time', 'f8', ('time',))
        time_var.standard_name = 'time'
        time_var.units = 'seconds since {}'.format(reference_date)
        time_var.axis = 'T'
        time_var[:] = times

    def write_ensemble_data(self, ensemble_size):

        if ensemble_size > 1:
            self.__ensemble_member_dim = self.__dataset.createDimension('realization', ensemble_size)
            ensemble_member_var = self.__dataset.createVariable('realization', 'i', ('realization',))
            ensemble_member_var.standard_name = 'realization'
            ensemble_member_var.long_name = 'Index of an ensemble member within an ensemble'
            ensemble_member_var.units = 1

    def write_station_data(self, stations: Stations, output_station_ids: List[str]) -> None:
        """
        Writes the station ids and additional station information to the given dataset.

        :param stations:           The stations data read from the input file.
        :param output_station_ids: The set of station ids for which output will be written. Must be unique.
        """
        assert len(set(output_station_ids)) == len(output_station_ids)

        self.__station_dim = self.__dataset.createDimension('station', len(output_station_ids))

        # first write the ids
        max_id_length = max(len(id) for id in output_station_ids)
        self.__dataset.createDimension('char_leng_id', max_id_length)
        station_id_var = self.__dataset.createVariable('station_id', 'c', ('station', 'char_leng_id'))
        station_id_var.long_name = 'station identification code'
        station_id_var.cf_role = 'timeseries_id'

        # we must store the index we use for each station id, to be able to write the data at the correct index later
        self.__station_id_to_index_mapping = {}
        for i, id in enumerate(output_station_ids):
            station_id_var[i, :] = list(id)
            self.__station_id_to_index_mapping[id] = i

        # now write the stored attributes
        for var_name, attr_var in stations.attribute_variables.items():
            variable = self.__dataset.createVariable(var_name, attr_var.datatype, ('station',))
            # copy all attributes from the original input variable
            variable.setncatts(attr_var.__dict__)

            for station_id in output_station_ids:
                if station_id in stations.attributes:
                    station_index = self.__station_id_to_index_mapping[station_id]
                    variable[station_index] = stations.attributes[station_id][var_name]

    def create_variables(self, variable_names: List[str], ensemble_size: int) -> None:
        """
        Creates variables in the dataset for each of the provided parameter ids.
        The write_times and write_station_data methods must be called first, to ensure the necessary dimensions have
        already been created in the output NetCDF file.

        :param variable_names: The parameter ids for which variables must be created. Must be unique.
        :param ensemble_size: the number of members in the ensemble
        """
        assert len(set(variable_names)) == len(variable_names)

        assert self.__time_dim is not None, 'First call write_times to ensure the time dimension has been created.'
        assert self.__station_dim is not None, 'First call write_station_data to ensure ' \
                                               'the station dimension has been created'
        assert self.__station_id_to_index_mapping is not None  # should also be created in write_station_data

        if ensemble_size > 1:
            assert self.__ensemble_member_dim is not None, \
                'First call write_ensemble_data to ensure the realization dimension has been created.'

            for variable_name in variable_names:
                self.__dataset.createVariable(variable_name,
                                              'f8', ('time', 'station', 'realization'), fill_value=np.nan)
        else:
            for variable_name in variable_names:
                self.__dataset.createVariable(variable_name,
                                              'f8', ('time', 'station'), fill_value=np.nan)

    def write_output_values(self, station_id: str, variable_name: str,
                            ensemble_member_index: int, values: np.ndarray,
                            ensemble_size: int) -> None:
        """
        Writes the given data to the dataset. The variable must have already been created through the
        create_variables method. After all calls to write_output_values, the close method must be called to flush all
        changes.

        :param station_id: The id of the station the data is written for.
        :param variable_name: The name of the variable the data is written to (must have already been created).
        :param ensemble_member_index: The index associated to the ensemble member
        :param values:        The values that are to be written to the file
        :param ensemble_size: the number of members in the ensemble
        """
        assert self.__station_id_to_index_mapping is not None, 'First call write_station_data and create_variables.'

        station_index = self.__station_id_to_index_mapping[station_id]
        if ensemble_size > 1:
            self.__dataset.variables[variable_name][:, station_index, ensemble_member_index] = values
        else:
            self.__dataset.variables[variable_name][:, station_index] = values

    def close(self) -> None:
        """
        Closes the NetCDF4 Dataset to ensure all changes made are written to the file.
        This method must be called after writing all data through the various write method.
        """
        self.__dataset.close()



from abc import ABCMeta

from casadi import if_else, logic_and


class BSpline(metaclass=ABCMeta):
    """
    B-Spline base class.
    """

    def basis(self, t, x, k, i):
        """
        Evaluate the B-Spline basis function using Cox-de Boor recursion.

        :param x: Point at which to evaluate.
        :param k: Order of the basis function.
        :param i: Knot number.

        :returns: The B-Spline basis function of the given order, at the given knot, evaluated at the given point.
        """
        if k == 0:
            return if_else(logic_and(t[i] <= x, x < t[i + 1]), 1.0, 0.0)
        else:
            if t[i] < t[i + k]:
                a = (x - t[i]) / (t[i + k] - t[i]) * self.basis(t, x, k - 1, i)
            else:
                a = 0.0
            if t[i + 1] < t[i + k + 1]:
                b = (t[i + k + 1] - x) / (t[i + k + 1] - t[i + 1]) * \
                    self.basis(t, x, k - 1, i + 1)
            else:
                b = 0.0
            return a + b



import bisect
import datetime
import io
import logging
import os
import xml.etree.ElementTree as ET

import numpy as np

ns = {'fews': 'http://www.wldelft.nl/fews',
      'pi': 'http://www.wldelft.nl/fews/PI'}


class Diag:
    """
    diag wrapper.
    """

    ERROR_FATAL = 1 << 0
    ERROR = 1 << 1
    WARN = 1 << 2
    INFO = 1 << 3
    DEBUG = 1 << 4

    def __init__(self, folder, basename='diag'):
        """
        Parse diag file.

        :param folder:   Folder in which diag.xml is found or to be created.
        :param basename: Alternative basename for the diagnostics XML file.
        """
        self.__path_xml = os.path.join(folder, basename + '.xml')

        self.__tree = ET.parse(self.__path_xml)
        self.__xml_root = self.__tree.getroot()

    def get(self, level=ERROR_FATAL):
        """
        Return only the wanted levels (debug, info, etc.)

        :param level: Log level.
        """

        diag_lines = self.__xml_root.findall("*", ns)
        diag_lines_out = []
        USED_LEVELS = []

        if level & self.ERROR_FATAL:
            USED_LEVELS.append('0')
        if level & self.ERROR:
            USED_LEVELS.append('1')
        if level & self.WARN:
            USED_LEVELS.append('2')
        if level & self.INFO:
            USED_LEVELS.append('3')
        if level & self.DEBUG:
            USED_LEVELS.append('4')

        for child in diag_lines:
            for used_level in USED_LEVELS:
                if child.get('level') == used_level:
                    diag_lines_out.append(child)

        return diag_lines_out

    @property
    def has_errors(self):
        """
        True if the log contains errors.
        """
        error_levels = self.ERROR_FATAL | self.ERROR
        diag_lines = self.get(error_levels)

        if len(diag_lines) > 0:
            return True
        else:
            return False


class DiagHandler(logging.Handler):
    """
    PI diag file logging handler.
    """

    def __init__(self, folder, basename='diag', level=logging.NOTSET):
        super(DiagHandler,  self).__init__(level=level)

        self.__path_xml = os.path.join(folder, basename + '.xml')

        try:
            self.__tree = ET.parse(self.__path_xml)
            self.__xml_root = self.__tree.getroot()
        except Exception:
            self.__xml_root = ET.Element('{%s}Diag' % (ns['pi'], ))
            self.__tree = ET.ElementTree(element=self.__xml_root)

        self.__map_level = {50: 0, 40: 1, 30: 2, 20: 3, 10: 4, 0: 4}

    def emit(self, record):
        self.format(record)

        self.acquire()
        el = ET.SubElement(self.__xml_root, '{%s}line' % (ns['pi'], ))
        # Work around cElementTree issue 21403
        el.set('description', record.message)
        el.set('eventCode', record.module + '.' + record.funcName)
        el.set('level', str(self.__map_level[record.levelno]))
        self.release()

    def append_element(self, el):
        self.acquire()
        self.__xml_root.append(el)
        self.release()

    def flush(self):
        self.__tree.write(self.__path_xml)

    def close(self):
        self.flush()
        super().close()


class ParameterConfig:
    """
    rtcParameterConfig wrapper.
    """

    def __init__(self, folder, basename):
        """
        Parses a rtcParameterConfig file.

        :param folder:    Folder in which the parameter configuration file is located.
        :param basename:  Basename of the parameter configuration file (e.g, 'rtcParameterConfig').
        """
        if os.path.splitext(basename)[1] != '.xml':
            basename = basename + '.xml'
        self.__path_xml = os.path.join(folder, basename)

        self.__tree = ET.parse(self.__path_xml)
        self.__xml_root = self.__tree.getroot()

    def get(self, group_id, parameter_id, location_id=None, model=None):
        """
        Returns the value of the parameter with ID parameter_id in the group with ID group_id.

        :param group_id:     The ID of the parameter group to look in.
        :param parameter_id: The ID of the parameter to look for.
        :param location_id:  The optional  ID of the parameter location to look in.
        :param model:        The optional ID of the parameter model to look in.

        :returns: The value of the specified parameter.
        """
        groups = self.__xml_root.findall(
            "pi:group[@id='{}']".format(group_id), ns)
        for group in groups:
            el = group.find("pi:locationId", ns)
            if location_id is not None and el is not None:
                if location_id != el.text:
                    continue

            el = group.find("pi:model", ns)
            if model is not None and el is not None:
                if model != el.text:
                    continue

            el = group.find("pi:parameter[@id='{}']".format(parameter_id), ns)
            if el is None:
                raise KeyError
            return self.__parse_parameter(el)

        raise KeyError("No such parameter ({}, {})".format(
            group_id, parameter_id))

    def set(self, group_id, parameter_id, new_value, location_id=None, model=None):
        """
        Set the value of the parameter with ID parameter_id in the group with ID group_id.

        :param group_id:     The ID of the parameter group to look in.
        :param parameter_id: The ID of the parameter to look for.
        :param new_value:    The new value for the parameter.
        :param location_id:  The optional  ID of the parameter location to look in.
        :param model:        The optional ID of the parameter model to look in.
        """
        groups = self.__xml_root.findall(
            "pi:group[@id='{}']".format(group_id), ns)
        for group in groups:
            el = group.find("pi:locationId", ns)
            if location_id is not None and el is not None:
                if location_id != el.text:
                    continue

            el = group.find("pi:model", ns)
            if model is not None and el is not None:
                if model != el.text:
                    continue

            el = group.find("pi:parameter[@id='{}']".format(parameter_id), ns)
            if el is None:
                raise KeyError
            for child in el:
                if child.tag.endswith('boolValue'):
                    if new_value is True:
                        child.text = 'true'
                        return
                    elif new_value is False:
                        child.text = 'false'
                        return
                    else:
                        raise Exception(
                            "Unsupported value for tag {}".format(child.tag))
                elif child.tag.endswith('intValue'):
                    child.text = str(int(new_value))
                    return
                elif child.tag.endswith('dblValue'):
                    child.text = str(new_value)
                    return
                else:
                    raise Exception("Unsupported tag {}".format(child.tag))

        raise KeyError("No such parameter ({}, {})".format(
            group_id, parameter_id))

    def write(self, folder=None, basename=None):
        """
        Writes the parameter configuration to a file.
        Default behaviour is to overwrite original file.

        :param path:     Optional alternative destination folder
        :param basename: Optional alternative basename of the file (e.g, 'rtcParameterConfig')
        """

        # No path changes- overwrite original file
        if folder is None and basename is None:
            path = self.path

        # We need to reconstruct the path
        else:
            # Determine folder
            if folder is not None:
                if not os.path.exists(folder):
                    # Make sure folder exists
                    raise FileNotFoundError('Folder not found: {}'.format(folder))
            else:
                # Reuse folder of original file
                folder = os.path.dirname(self.path)

            # Determine basename
            if basename is not None:
                # Make sure basename ends in '.xml'
                if os.path.splitext(basename)[1] != '.xml':
                    basename = basename + '.xml'
            else:
                # Reuse basename of original file
                basename = os.path.split(self.path)[1]

            # Construct path
            path = os.path.join(folder, basename)

        self.__tree.write(path)

    @property
    def path(self):
        return self.__path_xml

    def __parse_type(self, fews_type):
        # Parse a FEWS type to an np type
        if fews_type == 'double':
            return np.dtype('float')
        else:
            return np.dtype('S128')

    def __parse_parameter(self, parameter):
        for child in parameter:
            if child.tag.endswith('boolValue'):
                if child.text.lower() == 'true':
                    return True
                else:
                    return False
            elif child.tag.endswith('intValue'):
                return int(child.text)
            elif child.tag.endswith('dblValue'):
                return float(child.text)
            elif child.tag.endswith('stringValue'):
                return child.text
            # return dict of lisstart_datetime
            elif child.tag.endswith('table'):
                columnId = {}
                columnType = {}
                for key in child.find("pi:row", ns).attrib:
                    # default Id
                    columnId[key] = key
                    columnType[key] = np.dtype(
                        'S128')                     # default Type

                # get Id's if present
                el_columnIds = child.find("pi:columnIds", ns)
                if el_columnIds is not None:
                    for key, value in el_columnIds.attrib.items():
                        columnId[key] = value

                # get Types if present
                el_columnTypes = child.find("pi:columnTypes", ns)
                if el_columnTypes is not None:
                    for key, value in el_columnTypes.attrib.items():
                        columnType[key] = self.__parse_type(value)

                # get table contenstart_datetime
                el_row = child.findall("pi:row", ns)
                table = {columnId[key]: np.empty(len(el_row),            # initialize table
                                                 columnType[key]) for key in columnId}

                i_row = 0
                for row in el_row:
                    for key, value in row.attrib.items():
                        table[columnId[key]][i_row] = value
                    i_row += 1
                return table
            elif child.tag.endswith('description'):
                pass
            else:
                raise Exception("Unsupported tag {}".format(child.tag))

    def __iter__(self):
        # Iterate over all parameter key, value pairs.
        groups = self.__xml_root.findall("pi:group", ns)
        for group in groups:
            el = group.find("pi:locationId", ns)
            if el is not None:
                location_id = el.text
            else:
                location_id = None

            el = group.find("pi:model", ns)
            if el is not None:
                model_id = el.text
            else:
                model_id = None

            parameters = group.findall("pi:parameter", ns)
            for parameter in parameters:
                yield location_id, model_id, parameter.attrib['id'], self.__parse_parameter(parameter)


class Timeseries:
    """
    PI timeseries wrapper.
    """

    def __init__(self, data_config, folder, basename, binary=True, pi_validate_times=False, make_new_file=False):
        """
        Load the timeseries from disk.

        :param data_config:       A :class:`DataConfig` object.
        :param folder:            The folder in which the time series is located.
        :param basename:          The basename of the time series file.
        :param binary:            True if the time series data is stored in a separate binary file.
                                  Default is ``True``.
        :param pi_validate_times: Check consistency of times.  Default is ``False``.
        :param make_new_file:     Make new XML object which can be filled and written to a new file.
                                  Default is ``False``.
        """
        self.__data_config = data_config

        self.__folder = folder
        self.__basename = basename

        self.__path_xml = os.path.join(self.__folder, basename + '.xml')

        self.__internal_dtype = np.float64
        self.__pi_dtype = np.float32

        self.make_new_file = make_new_file
        if self.make_new_file:
            self.__reset_xml_tree()
        else:
            self.__tree = ET.parse(self.__path_xml)
            self.__xml_root = self.__tree.getroot()

        self.__values = [{}]
        self.__units = [{}]

        self.__binary = binary

        if not self.make_new_file:
            f = None
            if self.__binary:
                try:
                    f = io.open(self.binary_path, 'rb')
                except IOError:
                    # Support placeholder XML files.
                    pass

            # Read timezone
            timezone = self.__xml_root.find('pi:timeZone', ns)
            if timezone is not None:
                self.__timezone = float(timezone.text)
            else:
                self.__timezone = None

            # Check data consistency
            self.__dt = None
            self.__start_datetime = None
            self.__end_datetime = None
            self.__forecast_datetime = None
            self.__forecast_index = None
            self.__contains_ensemble = False
            self.__ensemble_size = 1
            for series in self.__xml_root.findall('pi:series', ns):
                header = series.find('pi:header', ns)

                variable = self.__data_config.variable(header)

                try:
                    dt = self.__parse_time_step(header.find('pi:timeStep', ns))
                except ValueError:
                    raise Exception(
                        'PI: Multiplier of time step of variable {} '
                        'must be a positive integer per the PI schema.'.format(variable))
                if self.__dt is None:
                    self.__dt = dt
                else:
                    if dt != self.__dt:
                        raise Exception(
                            'PI: Not all timeseries have the same time step size.')
                try:
                    start_datetime = self.__parse_date_time(
                        header.find('pi:startDate', ns))
                    if self.__start_datetime is None:
                        self.__start_datetime = start_datetime
                    else:
                        if start_datetime < self.__start_datetime:
                            self.__start_datetime = start_datetime
                except (AttributeError, ValueError):
                    raise Exception('PI: Variable {} in {} has no startDate.'.format(
                        variable, os.path.join(self.__folder, basename + '.xml')))

                try:
                    end_datetime = self.__parse_date_time(header.find('pi:endDate', ns))
                    if self.__end_datetime is None:
                        self.__end_datetime = end_datetime
                    else:
                        if end_datetime > self.__end_datetime:
                            self.__end_datetime = end_datetime
                except (AttributeError, ValueError):
                    raise Exception('PI: Variable {} in {} has no endDate.'.format(
                        variable, os.path.join(self.__folder, basename + '.xml')))

                el = header.find('pi:forecastDate', ns)
                if el is not None:
                    forecast_datetime = self.__parse_date_time(el)
                else:
                    # the timeseries has no forecastDate, so the forecastDate
                    # is set to the startDate (per the PI-schema)
                    forecast_datetime = start_datetime
                if self.__forecast_datetime is None:
                    self.__forecast_datetime = forecast_datetime
                else:
                    if el is not None and forecast_datetime != self.__forecast_datetime:
                        raise Exception(
                            'PI: Not all timeseries share the same forecastDate.')

                el = header.find('pi:ensembleMemberIndex', ns)
                if el is not None:
                    contains_ensemble = True
                    if int(el.text) > self.__ensemble_size - 1:  # Assume zero-based
                        self.__ensemble_size = int(el.text) + 1
                else:
                    contains_ensemble = False
                if self.__contains_ensemble is False:
                    # Only overwrite when _contains_ensemble was False before
                    self.__contains_ensemble = contains_ensemble

            # Define the times, and floor the global forecast_datetime to the
            # global time step to get its index
            if self.__dt:
                t_len = int(round(
                    (self.__end_datetime - self.__start_datetime).total_seconds() / self.__dt.total_seconds() + 1))
                self.__times = [self.__start_datetime + i * self.__dt for i in range(t_len)]
            else:  # Timeseries are non-equidistant
                self.__times = []
                for series in self.__xml_root.findall('pi:series', ns):
                    events = series.findall('pi:event', ns)
                    # We assume that timeseries can differ in length, but always are
                    # a complete 'slice' of datetimes between start and end. The
                    # longest timeseries then contains all datetimes between start and end.
                    if len(events) > len(self.__times):
                        self.__times = [self.__parse_date_time(e) for e in events]

            # Check if the time steps of all series match the time steps of the global
            # time range.
            if pi_validate_times:
                ref_times_set = set(self.__times)
                for series in self.__xml_root.findall('pi:series', ns):
                    events = series.findall('pi:event', ns)
                    times = {self.__parse_date_time(e) for e in events}
                    if not ref_times_set.issuperset(times):
                        raise ValueError(
                            'PI: Not all timeseries share the same time step spacing. Make sure '
                            'the time steps of all series are a subset of the global time steps.')

            if self.__forecast_datetime is not None:
                if self.__dt:
                    self.__forecast_datetime = self.__floor_date_time(
                        dt=self.__forecast_datetime, tdel=self.__dt)
                try:
                    self.__forecast_index = self.__times.index(
                        self.__forecast_datetime)
                except ValueError:
                    # This may occur if forecast_datetime is outside of
                    # the timeseries' range.  Can be a valid case for historical
                    # timeseries, for instance.
                    self.__forecast_index = -1

            # Parse data
            for series in self.__xml_root.findall('pi:series', ns):
                header = series.find('pi:header', ns)

                variable = self.__data_config.variable(header)

                dt = self.__parse_time_step(header.find('pi:timeStep', ns))
                start_datetime = self.__parse_date_time(header.find('pi:startDate', ns))
                end_datetime = self.__parse_date_time(header.find('pi:endDate', ns))

                make_virtual_ensemble = False
                el = header.find('pi:ensembleMemberIndex', ns)
                if el is not None:
                    ensemble_member = int(el.text)
                    while ensemble_member >= len(self.__values):
                        self.__values.append({})
                    while ensemble_member >= len(self.__units):
                        self.__units.append({})
                else:
                    ensemble_member = 0
                if el is None and self.contains_ensemble is True:
                    # Expand values dict to accommodate referencing of (virtual)
                    # ensemble series to the input values. This is e.g. needed
                    # for initial states that have a single historical values.
                    while self.ensemble_size > len(self.__values):
                        self.__values.append({})
                    while self.ensemble_size > len(self.__units):
                        self.__units.append({})
                    make_virtual_ensemble = True

                if self.__dt:
                    n_values = int(
                        round((end_datetime - start_datetime).total_seconds() / dt.total_seconds() + 1))
                else:
                    n_values = (bisect.bisect_left(self.__times, end_datetime) -
                                bisect.bisect_left(self.__times, start_datetime) + 1)

                if self.__binary:
                    if f is not None:
                        self.__values[ensemble_member][variable] = np.fromfile(
                            f, count=n_values, dtype=self.__pi_dtype)
                    else:
                        self.__values[ensemble_member][variable] = np.empty(
                            n_values, dtype=self.__internal_dtype)
                        self.__values[ensemble_member][variable].fill(np.nan)

                else:
                    events = series.findall('pi:event', ns)

                    self.__values[ensemble_member][variable] = np.empty(
                        n_values, dtype=self.__internal_dtype)
                    self.__values[ensemble_member][variable].fill(np.nan)
                    # This assumes that start_datetime equals the datetime of the
                    # first value (which should be the case).
                    for i in range(min(n_values, len(events))):
                        self.__values[ensemble_member][variable][
                            i] = float(events[i].get('value'))

                miss_val = float(header.find('pi:missVal', ns).text)
                self.__values[ensemble_member][variable][self.__values[
                    ensemble_member][variable] == miss_val] = np.nan

                unit = header.find('pi:units', ns).text
                self.set_unit(variable, unit=unit, ensemble_member=ensemble_member)

                # Prepend empty space, if start_datetime > self.__start_datetime.
                if start_datetime > self.__start_datetime:
                    if self.__dt:
                        filler = np.empty(
                            int(round((start_datetime - self.__start_datetime).total_seconds() / dt.total_seconds())),
                            dtype=self.__internal_dtype)
                    else:
                        filler = np.empty(
                            int(round(bisect.bisect_left(self.__times, start_datetime) -
                                      bisect.bisect_left(self.__times, self.__start_datetime))),
                            dtype=self.__internal_dtype)

                    filler.fill(np.nan)
                    self.__values[ensemble_member][variable] = np.hstack(
                        (filler, self.__values[ensemble_member][variable]))

                # Append empty space, if end_datetime < self.__end_datetime
                if end_datetime < self.__end_datetime:
                    if self.__dt:
                        filler = np.empty(
                            int(round((self.__end_datetime - end_datetime).total_seconds() / dt.total_seconds())),
                            dtype=self.__internal_dtype)
                    else:
                        filler = np.empty(
                            int(round(bisect.bisect_left(self.__times, self.__end_datetime) -
                                      bisect.bisect_left(self.__times, end_datetime))),
                            dtype=self.__internal_dtype)

                    filler.fill(np.nan)
                    self.__values[ensemble_member][variable] = np.hstack(
                        (self.__values[ensemble_member][variable], filler))

                if make_virtual_ensemble:
                    # Make references to the original input series for the virtual
                    # ensemble members.
                    for i in range(1, self.ensemble_size):
                        self.__values[i][variable] = self.__values[0][variable]
                        self.set_unit(variable, unit=unit, ensemble_member=i)

            if not self.__dt:
                # Remove time values outside the start/end datetimes.
                # Only needed for non-equidistant, because we can't build the
                # times automatically from global start/end datetime.
                self.__times = self.__times[bisect.bisect_left(self.__times, self.__start_datetime):
                                            bisect.bisect_left(self.__times, self.__end_datetime) + 1]

            if f is not None and self.__binary:
                f.close()

    def __reset_xml_tree(self):
        # Make a new empty XML tree
        self.__xml_root = ET.Element('{%s}' % (ns['pi'], ) + 'TimeSeries')
        self.__tree = ET.ElementTree(self.__xml_root)

        self.__xml_root.set('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
        self.__xml_root.set('version', '1.2')
        self.__xml_root.set(
            'xsi:schemaLocation',
            'http://www.wldelft.nl/fews/PI http://fews.wldelft.nl/schemas/version1.0/pi-schemas/pi_timeseries.xsd')

    def __parse_date_time(self, el):
        # Parse a PI date time element.
        return datetime.datetime.strptime(el.get('date') + ' ' + el.get('time'), '%Y-%m-%d %H:%M:%S')

    def __parse_time_step(self, el):
        # Parse a PI time step element.
        if el.get('unit') == 'second':
            return datetime.timedelta(seconds=int(el.get('multiplier')))
        elif el.get('unit') == 'nonequidistant':
            return None
        else:
            raise Exception('Unsupported unit type: ' + el.get('unit'))

    def __floor_date_time(self, dt, tdel):
        """
        Floor a PI date time to an integer number of PI time steps from the
        start date time.
        """
        roundTo = tdel.total_seconds()

        seconds = (dt - self.__start_datetime).seconds
        # // is a floor division:
        rounding = (seconds + roundTo / 2) // roundTo * roundTo
        return dt + datetime.timedelta(0, rounding - seconds, -dt.microsecond)

    def __add_header(self, variable, location_parameter_id, ensemble_member=0, miss_val=-999, unit='unit_unknown'):
        """
        Add a timeseries header to the timeseries object.
        """
        # Save current datetime
        now = datetime.datetime.now()

        # Define the basic structure of the header
        header_elements = [
            'type', 'locationId', 'parameterId', 'timeStep', 'startDate', 'endDate',
            'missVal', 'stationName', 'units', 'creationDate', 'creationTime']
        header_element_texts = [
            'instantaneous', location_parameter_id.location_id, location_parameter_id.parameter_id,
            '', '', '', str(miss_val), location_parameter_id.location_id, unit,
            now.strftime('%Y-%m-%d'), now.strftime('%H:%M:%S')]

        # Add ensembleMemberIndex, forecastDate and qualifierId if necessary.
        if self.__forecast_datetime != self.__start_datetime:
            header_elements.insert(6, 'forecastDate')
            header_element_texts.insert(6, '')
        if self.contains_ensemble:
            header_elements.insert(3, 'ensembleMemberIndex')
            header_element_texts.insert(3, str(ensemble_member))
        if len(location_parameter_id.qualifier_id) > 0:
            # Track relative index to preserve original ordering of qualifier ID's
            i = 0
            for qualifier_id in location_parameter_id.qualifier_id:
                header_elements.insert(3, 'qualifierId')
                header_element_texts.insert(3+i, qualifier_id)
                i += 1

        # Fill the basics of the series
        series = ET.Element('{%s}' % (ns['pi'], ) + 'series')
        header = ET.SubElement(series, '{%s}' % (ns['pi'], ) + 'header')
        for i in range(len(header_elements)):
            el = ET.SubElement(header, '{%s}' % (ns['pi'], ) + header_elements[i])
            el.text = header_element_texts[i]

        el = header.find('pi:timeStep', ns)
        # Set time step
        if self.dt:
            el.set('unit', 'second')
            el.set('multiplier', str(int(self.dt.total_seconds())))
        else:
            el.set('unit', 'nonequidistant')

        # Set the time range.
        el = header.find('pi:startDate', ns)
        el.set('date', self.__start_datetime.strftime('%Y-%m-%d'))
        el.set('time', self.__start_datetime.strftime('%H:%M:%S'))
        el = header.find('pi:endDate', ns)
        el.set('date', self.__end_datetime.strftime('%Y-%m-%d'))
        el.set('time', self.__end_datetime.strftime('%H:%M:%S'))

        # Set the forecast date if applicable
        if self.__forecast_datetime != self.__start_datetime:
            el = header.find('pi:forecastDate', ns)
            el.set('date', self.__forecast_datetime.strftime('%Y-%m-%d'))
            el.set('time', self.__forecast_datetime.strftime('%H:%M:%S'))

        # Add series to xml
        self.__xml_root.append(series)

    def write(self):
        """
        Writes the time series data to disk.
        """

        if self.__binary:
            f = io.open(self.binary_path, 'wb')

        if self.timezone is not None:
            timezone = self.__xml_root.find('pi:timeZone', ns)
            if timezone is None:
                timezone = ET.Element('{%s}' % (ns['pi'], ) + 'timeZone')
                # timeZone has to be the first element according to the schema
                self.__xml_root.insert(0, timezone)
            timezone.text = str(self.timezone)

        if self.make_new_file:
            # Force reinitialization in case write() is called more than once
            self.__reset_xml_tree()

            for ensemble_member in range(len(self.__values)):
                for variable in sorted(self.__values[ensemble_member].keys()):
                    location_parameter_id = self.__data_config.pi_variable_ids(variable)
                    unit = self.get_unit(variable, ensemble_member)
                    self.__add_header(
                        variable, location_parameter_id,
                        ensemble_member=ensemble_member, miss_val=-999, unit=unit)

        for ensemble_member in range(len(self.__values)):
            for series in self.__xml_root.findall('pi:series', ns):
                header = series.find('pi:header', ns)

                # First check ensembleMemberIndex, to see if it is the correct one.
                el = header.find('pi:ensembleMemberIndex', ns)
                if el is not None:
                    if ensemble_member != int(el.text):
                        # Skip over this series, wrong index.
                        continue

                # Update the time range, which may have changed.
                el = header.find('pi:startDate', ns)
                el.set('date', self.__start_datetime.strftime('%Y-%m-%d'))
                el.set('time', self.__start_datetime.strftime('%H:%M:%S'))

                el = header.find('pi:endDate', ns)
                el.set('date', self.__end_datetime.strftime('%Y-%m-%d'))
                el.set('time', self.__end_datetime.strftime('%H:%M:%S'))

                variable = self.__data_config.variable(header)

                miss_val = float(header.find('pi:missVal', ns).text)
                values = self.__values[ensemble_member][variable]

                # Update the header, which may have changed
                el = header.find('pi:units', ns)
                el.text = self.get_unit(variable, ensemble_member)

                # No values to be written, so the entire element is removed from
                # the XML, and the loop restarts.
                if len(values) == 0:
                    self.__xml_root.remove(series)
                    continue

                # Replace NaN with missing value
                nans = np.isnan(values)
                values[nans] = miss_val

                # Write output
                if self.__binary:
                    f.write(values.astype(self.__pi_dtype).tostring())
                else:
                    events = series.findall('pi:event', ns)

                    t = self.__start_datetime
                    for i in range(min(len(events), len(values))):
                        if self.dt is None:
                            t = self.times[i]
                        # Set the date/time, so that any date/time steps that
                        # are wrong in the placeholder file are corrected.
                        events[i].set('date', t.strftime('%Y-%m-%d'))
                        events[i].set('time', t.strftime('%H:%M:%S'))

                        # Set the value
                        events[i].set('value', str(values[i]))
                        if self.dt:
                            t += self.dt
                    for i in range(len(events), len(values)):
                        if self.dt is None:
                            t = self.times[i]
                        event = ET.Element('pi:event')
                        event.set('date', t.strftime('%Y-%m-%d'))
                        event.set('time', t.strftime('%H:%M:%S'))
                        event.set('value', str(values[i]))
                        series.append(event)
                        if self.dt:
                            t += self.dt

                    # Remove superfluous elements
                    if len(events) > len(values):
                        for i in range(len(values), len(events)):
                            series.remove(events[i])

                # Restore NaN
                values[nans] = np.nan

        if self.__binary:
            f.close()

        self.__tree.write(self.__path_xml)

    @property
    def contains_ensemble(self):
        """
        Flag to indicate TimeSeries contains an ensemble.
        """
        return self.__contains_ensemble

    @contains_ensemble.setter
    def contains_ensemble(self, contains_ensemble):
        if not contains_ensemble:
            assert self.ensemble_size == 1
        self.__contains_ensemble = contains_ensemble

    @property
    def ensemble_size(self):
        """
        Ensemble size.
        """
        return self.__ensemble_size

    @ensemble_size.setter
    def ensemble_size(self, ensemble_size):
        if ensemble_size > 1:
            assert self.contains_ensemble
        self.__ensemble_size = ensemble_size
        while self.__ensemble_size > len(self.__values):
            self.__values.append({})
        while self.__ensemble_size > len(self.__units):
            self.__units.append({})

    @property
    def start_datetime(self):
        """
        Start time.
        """
        return self.__start_datetime

    @property
    def end_datetime(self):
        """
        End time.
        """
        return self.__end_datetime

    @property
    def forecast_datetime(self):
        """
        Forecast time (t0).
        """
        return self.__forecast_datetime

    @forecast_datetime.setter
    def forecast_datetime(self, forecast_datetime):
        self.__forecast_datetime = forecast_datetime
        self.__forecast_index = self.__times.index(forecast_datetime)

    @property
    def forecast_index(self):
        """
        Forecast time (t0) index.
        """
        return self.__forecast_index

    @property
    def dt(self):
        """
        Time step.
        """
        return self.__dt

    @dt.setter
    def dt(self, dt):
        self.__dt = dt

    @property
    def times(self):
        """
        Time stamps.
        """
        return self.__times

    @times.setter
    def times(self, times):
        self.__times = times
        self.__start_datetime = times[0]
        self.__end_datetime = times[-1]

    @property
    def timezone(self):
        """
        Time zone in decimal hours shift from GMT.
        """
        return self.__timezone

    @timezone.setter
    def timezone(self, timezone):
        self.__timezone = timezone

    def get(self, variable, ensemble_member=0):
        """
        Look up a time series.

        :param variable:        Time series ID.
        :param ensemble_member: Ensemble member index.

        :returns: A :class:`Timeseries` object.
        """
        return self.__values[ensemble_member][variable]

    def set(self, variable, new_values, unit=None, ensemble_member=0):
        """
        Fill a time series with new values, and set the unit.

        :param variable:        Time series ID.
        :param new_values:      List of new values.
        :param unit:            Unit.
        :param ensemble_member: Ensemble member index.
        """
        self.__values[ensemble_member][variable] = new_values
        if unit is None:
            unit = self.get_unit(variable, ensemble_member)
        self.set_unit(variable, unit, ensemble_member)

    def get_unit(self, variable, ensemble_member=0):
        """
        Look up the unit of a time series.

        :param variable:        Time series ID.
        :param ensemble_member: Ensemble member index.

        :returns: A :string: containing the unit. If not set for the variable, returns 'unit_unknown'.
        """
        try:
            return self.__units[ensemble_member][variable]
        except KeyError:
            return 'unit_unknown'

    def set_unit(self, variable, unit, ensemble_member=0):
        """
        Set the unit of a time series.

        :param variable:        Time series ID.
        :param unit:            Unit.
        :param ensemble_member: Ensemble member index.
        """
        self.__units[ensemble_member][variable] = unit

    def resize(self, start_datetime, end_datetime):
        """
        Resize the timeseries to stretch from start_datetime to end_datetime.

        :param start_datetime: Start date and time.
        :param end_datetime:   End date and time.
        """

        if self.__dt:
            n_delta_s = int(round(
                (start_datetime - self.__start_datetime).total_seconds() / self.__dt.total_seconds()))
        else:
            if start_datetime >= self.__start_datetime:
                n_delta_s = bisect.bisect_left(self.__times, start_datetime) - \
                    bisect.bisect_left(self.__times, self.__start_datetime)
            else:
                raise ValueError(
                    'PI: Resizing a non-equidistant timeseries to stretch '
                    'outside of the global range of times is not allowed.')

        for ensemble_member in range(len(self.__values)):
            if n_delta_s > 0:
                # New start datetime lies after old start datetime (timeseries will be shortened).
                for key in self.__values[ensemble_member].keys():
                    self.__values[ensemble_member][key] = self.__values[
                        ensemble_member][key][n_delta_s:]
            elif n_delta_s < 0:
                # New start datetime lies before old start datetime (timeseries will be lengthened).
                filler = np.empty(abs(n_delta_s))
                filler.fill(np.nan)
                for key in self.__values[ensemble_member].keys():
                    self.__values[ensemble_member][key] = np.hstack(
                        (filler, self.__values[ensemble_member][key]))
        self.__start_datetime = start_datetime

        if self.__dt:
            n_delta_e = int(round(
                (end_datetime - self.__end_datetime).total_seconds() / self.__dt.total_seconds()))
        else:
            if end_datetime <= self.__end_datetime:
                n_delta_e = bisect.bisect_left(self.__times, end_datetime) - \
                    bisect.bisect_left(self.__times, self.__end_datetime)
            else:
                raise ValueError(
                    'PI: Resizing a non-equidistant timeseries to stretch '
                    'outside of the global range of times is not allowed.')

        for ensemble_member in range(len(self.__values)):
            if n_delta_e > 0:
                # New end datetime lies after old end datetime (timeseries will be lengthened).
                filler = np.empty(n_delta_e)
                filler.fill(np.nan)
                for key in self.__values[ensemble_member].keys():
                    self.__values[ensemble_member][key] = np.hstack(
                        (self.__values[ensemble_member][key], filler))
            elif n_delta_e < 0:
                # New end datetime lies before old end datetime (timeseries will be shortened).
                for key in self.__values[ensemble_member].keys():
                    self.__values[ensemble_member][key] = self.__values[
                        ensemble_member][key][:n_delta_e]
        self.__end_datetime = end_datetime

    @property
    def path(self):
        return self.__path_xml

    @property
    def binary_path(self):
        """
        The path for the binary data .bin file.
        """
        return os.path.join(self.__folder, self.__basename + '.bin')

    def items(self, ensemble_member=0):
        """
        Returns an iterator over all timeseries IDs and value arrays for the given
        ensemble member.
        """
        for key in self.__values[ensemble_member].keys():
            yield key, self.get(key, ensemble_member)



import logging
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta
from typing import Iterable, List, Tuple, Union

import numpy as np

from rtctools._internal.alias_tools import AliasDict, AliasRelation

logger = logging.getLogger("rtctools")


class DataStoreAccessor(metaclass=ABCMeta):
    """
    Base class for all problems.
    Adds an internal data store where timeseries and parameters can be stored.
    Access to the internal data store is always done through the io accessor.

    :cvar timeseries_import_basename:
        Import file basename. Default is ``timeseries_import``.
    :cvar timeseries_export_basename:
        Export file basename. Default is ``timeseries_export``.
    """

    #: Import file basename
    timeseries_import_basename = 'timeseries_import'
    #: Export file basename
    timeseries_export_basename = 'timeseries_export'

    def __init__(self, **kwargs):
        # Save arguments
        self._input_folder = kwargs['input_folder'] if 'input_folder' in kwargs else 'input'
        self._output_folder = kwargs['output_folder'] if 'output_folder' in kwargs else 'output'

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("Expecting input files to be located in '" + self._input_folder + "'.")
            logger.debug("Writing output files to '" + self._output_folder + "'.")

        self.io = DataStore(self)

    @property
    @abstractmethod
    def alias_relation(self) -> AliasRelation:
        raise NotImplementedError


class DataStore(metaclass=ABCMeta):
    """
    DataStore class used by the DataStoreAccessor.
    Contains all methods needed to access the internal data store.
    """

    def __init__(self, accessor):
        self.__accessor = accessor

        # Should all be set by subclass via setters
        self.__reference_datetime = None
        self.__timeseries_datetimes = None
        self.__timeseries_times_sec = None
        self.__timeseries_values = [AliasDict(self.__accessor.alias_relation)]
        self.__parameters = [AliasDict(self.__accessor.alias_relation)]

        self.__reference_datetime_fixed = False

        self.__ensemble_size = 1

    @property
    def reference_datetime(self):
        return self.__reference_datetime

    @reference_datetime.setter
    def reference_datetime(self, value):
        if self.__reference_datetime_fixed and value != self.__reference_datetime:
            raise RuntimeError("Cannot change reference datetime after times in seconds has been requested.")
        self.__reference_datetime = value

    @property
    def ensemble_size(self):
        return self.__ensemble_size

    def __update_ensemble_size(self, ensemble_size):
        while ensemble_size > len(self.__timeseries_values):
            self.__timeseries_values.append(AliasDict(self.__accessor.alias_relation))

        while ensemble_size > len(self.__parameters):
            self.__parameters.append(AliasDict(self.__accessor.alias_relation))

        assert len(self.__parameters) == len(self.__timeseries_values)
        assert len(self.__parameters) == ensemble_size

        self.__ensemble_size = ensemble_size

    @property
    def datetimes(self) -> List[datetime]:
        """"
        Returns the timeseries times in seconds.

        :returns: timeseries datetimes, or None if there has been no call
                  to :py:meth:`set_timeseries`.
        """
        return self.__timeseries_datetimes.copy()

    @property
    def times_sec(self) -> np.ndarray:
        """"
        Returns the timeseries times in seconds.

        Note that once this method is called, it is no longer allowed to
        change :py:attr:`reference_datetime`.

        :returns: timeseries times in seconds.
        """
        self._datetimes_to_seconds()

        return self.__timeseries_times_sec

    def _datetimes_to_seconds(self):
        if self.__reference_datetime_fixed:
            pass
        else:
            # Currently we only allow a reference datetime that exists in the
            # timeseries datetimes. That way we can guarantee that we have
            # "0.0" as one of our times in seconds. This restriction may be
            # loosened in the future.
            if self.reference_datetime not in self.__timeseries_datetimes:
                raise Exception("Reference datetime {} should be equal to one of the timeseries datetimes {}".format(
                    self.reference_datetime, self.__timeseries_datetimes))
            self.__timeseries_times_sec = self.datetime_to_sec(self.__timeseries_datetimes, self.reference_datetime)
            self.__timeseries_times_sec.flags.writeable = False
            self.__reference_datetime_fixed = True

    def set_timeseries(self,
                       variable: str,
                       datetimes: Iterable[datetime],
                       values: np.ndarray,
                       ensemble_member: int = 0,
                       check_duplicates: bool = False) -> None:
        """
        Stores input time series values in the internal data store.

        :param variable:         Variable name.
        :param datetimes:        Times as datetime objects.
        :param values:           The values to be stored.
        :param ensemble_member:  The ensemble member index.
        :param check_duplicates: If True, a warning will be given when overwriting values.
                                 If False, existing values can be silently overwritten with new values.
        """
        datetimes = list(datetimes)

        if not isinstance(datetimes[0], datetime):
            raise TypeError("DateStore.set_timeseries() only support datetimes")

        if self.__timeseries_datetimes is not None and datetimes != self.__timeseries_datetimes:
            raise RuntimeError("Attempting to overwrite the input time series datetimes with different values. "
                               "Please ensure all input time series have the same datetimes.")
        self.__timeseries_datetimes = datetimes

        if len(self.__timeseries_datetimes) != len(values):
            raise ValueError("Length of values ({}) must be the same as length of datetimes ({})"
                             .format(len(values), len(self.__timeseries_datetimes)))

        if ensemble_member >= self.__ensemble_size:
            self.__update_ensemble_size(ensemble_member + 1)

        if check_duplicates and variable in self.__timeseries_values[ensemble_member].keys():
            logger.warning("Time series values for ensemble member {} and variable {} set twice. "
                           "Overwriting old values.".format(ensemble_member, variable))

        self.__timeseries_values[ensemble_member][variable] = values

    def get_timeseries(self, variable: str, ensemble_member: int = 0) -> Tuple[List[datetime], np.ndarray]:
        """
        Looks up the time series in the internal data store.

        :return a tuple (datetimes, values)
        """
        if ensemble_member >= self.__ensemble_size:
            raise KeyError("ensemble_member {} does not exist".format(ensemble_member))
        return self.__timeseries_datetimes, self.__timeseries_values[ensemble_member][variable]

    def get_timeseries_names(self, ensemble_member: int = 0) -> Iterable[str]:
        return self.__timeseries_values[ensemble_member].keys()

    def set_timeseries_sec(self,
                           variable: str,
                           times_in_sec: np.ndarray,
                           values: np.ndarray,
                           ensemble_member: int = 0,
                           check_duplicates: bool = False) -> None:
        """
        Stores input time series values in the internal data store.

        Note that once this method is called, it is no longer allowed to
        change :py:attr:`reference_datetime`.

        :param variable:         Variable name.
        :param times_in_sec:     The times in seconds.
        :param values:           The values to be stored.
        :param ensemble_member:  The ensemble member index.
        :param check_duplicates: If True, a warning will be given when overwriting values.
                                 If False, existing values can be silently overwritten with new values.
        """
        self._datetimes_to_seconds()

        if self.reference_datetime is None:
            raise RuntimeError("Cannot use times in seconds before reference datetime is set.")

        if self.__timeseries_times_sec is not None and not np.array_equal(times_in_sec, self.__timeseries_times_sec):
            raise RuntimeError("Attempting to overwrite the input time series times with different values. "
                               "Please ensure all input time series have the same times.")

        if len(self.__timeseries_datetimes) != len(values):
            raise ValueError("Length of values ({}) must be the same as length of times ({})"
                             .format(len(values), len(self.__timeseries_datetimes)))

        if ensemble_member >= self.__ensemble_size:
            self.__update_ensemble_size(ensemble_member + 1)

        if check_duplicates and variable in self.__timeseries_values[ensemble_member].keys():
            logger.warning("Time series values for ensemble member {} and variable {} set twice. "
                           "Overwriting old values.".format(ensemble_member, variable))

        self.__timeseries_values[ensemble_member][variable] = values

    def get_timeseries_sec(self, variable: str, ensemble_member: int = 0) -> Tuple[np.ndarray, np.ndarray]:
        """
        Looks up the time series in the internal data store.

        Note that once this method is called, it is no longer allowed to
        change :py:attr:`reference_datetime`.

        :return a tuple (times, values)
        """
        self._datetimes_to_seconds()

        if ensemble_member >= self.__ensemble_size:
            raise KeyError("ensemble_member {} does not exist".format(ensemble_member))
        return self.__timeseries_times_sec, self.__timeseries_values[ensemble_member][variable]

    def set_parameter(self,
                      parameter_name: str,
                      value: float,
                      ensemble_member: int = 0,
                      check_duplicates: bool = False) -> None:
        """
        Stores the parameter value in the internal data store.

        :param parameter_name:   Parameter name.
        :param value:            The values to be stored.
        :param ensemble_member:  The ensemble member index.
        :param check_duplicates: If True, a warning will be given when overwriting values.
                                 If False, existing values can be silently overwritten with new values.
        """
        if ensemble_member >= self.__ensemble_size:
            self.__update_ensemble_size(ensemble_member + 1)

        if check_duplicates and parameter_name in self.__parameters[ensemble_member].keys():
            logger.warning("Attempting to set parameter value for ensemble member {} and name {} twice. "
                           "Using new value of {}.".format(ensemble_member, parameter_name, value))

        self.__parameters[ensemble_member][parameter_name] = value

    def get_parameter(self, parameter_name: str, ensemble_member: int = 0) -> float:
        """
        Looks up the parameter value in the internal data store.
        """
        if ensemble_member >= self.__ensemble_size:
            raise KeyError("ensemble_member {} does not exist".format(ensemble_member))
        return self.__parameters[ensemble_member][parameter_name]

    def parameters(self, ensemble_member: int = 0) -> AliasDict:
        """
        Returns an AliasDict of parameters to its values for the specified ensemble member.
        """
        if ensemble_member >= self.__ensemble_size:
            raise KeyError("ensemble_member {} does not exist".format(ensemble_member))
        return self.__parameters[ensemble_member]

    @staticmethod
    def datetime_to_sec(d: Union[Iterable[datetime], datetime], t0: datetime) -> Union[np.ndarray, float]:
        """
        Returns the date/timestamps in seconds since t0.

        :param d:  Iterable of datetimes or a single datetime object.
        :param t0: Reference datetime.
        """
        if hasattr(d, '__iter__'):
            return np.array([(t - t0).total_seconds() for t in d])
        else:
            return (d - t0).total_seconds()

    @staticmethod
    def sec_to_datetime(s: Union[Iterable[float], float], t0: datetime) -> Union[List[datetime], datetime]:
        """
        Return the date/timestamps in seconds since t0 as datetime objects.

        :param s:  Iterable of ints or a single int (number of seconds before or after t0).
        :param t0: Reference datetime.
        """
        if hasattr(s, '__iter__'):
            return [t0 + timedelta(seconds=t) for t in s]
        else:
            return t0 + timedelta(seconds=s)



import cProfile
import logging
import os
import pstats
import re
import sys

import casadi

from . import __version__
from ._internal.alias_tools import OrderedSet
from .data import pi
from .optimization.pi_mixin import PIMixin as OptimizationPIMixin
from .simulation.pi_mixin import PIMixin as SimulationPIMixin


def _resolve_folder(kwargs, base_folder, subfolder_kw, default):
    subfolder = kwargs.pop(subfolder_kw, default)
    if os.path.isabs(subfolder):
        return subfolder
    else:
        return os.path.join(base_folder, subfolder)


def run_optimization_problem(optimization_problem_class,
                             base_folder='..', log_level=logging.INFO, profile=False,
                             **kwargs):
    """
    Sets up and solves an optimization problem.

    This function makes the following assumptions:

    1. That the ``base_folder`` contains subfolders ``input``, ``output``, and ``model``,
       containing input data, output data, and the model, respectively.
    2. When using :class:`.CSVLookupTableMixin`, that the base folder contains a subfolder ``lookup_tables``.
    3. When using :class:`.ModelicaMixin`, that the base folder contains a subfolder ``model``.
    4. When using :class:`.ModelicaMixin`, that the toplevel Modelica model name equals the class name.

    :param optimization_problem_class: Optimization problem class to solve.
    :param base_folder:                Base folder.
    :param log_level:                  The log level to use.
    :param profile:                    Whether or not to enable profiling.

    :returns: :class:`.OptimizationProblem` instance.
    """

    if not os.path.isabs(base_folder):
        # Resolve base folder relative to script folder
        base_folder = os.path.join(sys.path[0], base_folder)

    model_folder = _resolve_folder(kwargs, base_folder, 'model_folder', 'model')
    input_folder = _resolve_folder(kwargs, base_folder, 'input_folder', 'input')
    output_folder = _resolve_folder(kwargs, base_folder, 'output_folder', 'output')

    # Set up logging
    logger = logging.getLogger("rtctools")

    # Add stream handler if it does not already exist.
    if not logger.hasHandlers() and not any((isinstance(h, logging.StreamHandler) for h in logger.handlers)):
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    # Add pi.DiagHandler, if using PIMixin. Only add it if it does not already exist.
    if (issubclass(optimization_problem_class, OptimizationPIMixin) and
            not any((isinstance(h, pi.DiagHandler) for h in logger.handlers))):
        handler = pi.DiagHandler(output_folder)
        logger.addHandler(handler)

    # Set log level
    logger.setLevel(log_level)

    # Log version info
    logger.info(
        "Using RTC-Tools {}.".format(__version__))
    logger.debug(
        "Using CasADi {}.".format(casadi.__version__))

    # Check for some common mistakes in inheritance order
    suggested_order = OrderedSet([
        'HomotopyMixin',
        'MinAbsGoalProgrammingMixin', 'LinearizedOrderGoalProgrammingMixin',
        'SinglePassGoalProgrammingMixin', 'GoalProgrammingMixin',
        'PIMixin', 'CSVMixin', 'ModelicaMixin',
        'ControlTreeMixin', 'CollocatedIntegratedOptimizationProblem', 'OptimizationProblem'])
    base_names = OrderedSet([b.__name__ for b in optimization_problem_class.__bases__])
    if suggested_order & base_names != base_names & suggested_order:
        msg = 'Please inherit from base classes in the following order: {}'.format(list(base_names & suggested_order))
        logger.error(msg)
        raise Exception(msg)

    # Run
    try:
        prob = optimization_problem_class(
            model_folder=model_folder, input_folder=input_folder, output_folder=output_folder,
            **kwargs)
        if profile:
            filename = os.path.join(base_folder, "profile.prof")

            cProfile.runctx("prob.optimize()", globals(), locals(), filename)

            s = pstats.Stats(filename)
            s.strip_dirs().sort_stats("time").print_stats()
        else:
            prob.optimize()
        return prob
    except Exception as e:
        logger.error(str(e))
        if isinstance(e, TypeError):
            exc_info = sys.exc_info()
            value = exc_info[1]
            try:
                failed_class = re.search(
                    "Can't instantiate (.*) with abstract methods", str(value)).group(1)
                abstract_method = re.search(
                    ' with abstract methods (.*)', str(value)).group(1)
                logger.error(
                    'The {} is missing a mixin. Please add a mixin that instantiates '
                    'abstract method {}, so that the optimizer can run.'.format(
                        failed_class, abstract_method))
            except Exception:
                pass
        for handler in logger.handlers:
            handler.flush()
        raise


def run_simulation_problem(simulation_problem_class,
                           base_folder='..', log_level=logging.INFO,
                           **kwargs):
    """
    Sets up and runs a simulation problem.

    :param simulation_problem_class: Optimization problem class to solve.
    :param base_folder:              Folder within which subfolders "input", "output", and "model" exist,
                                     containing input and output data, and the model, respectively.
    :param log_level:                The log level to use.

    :returns: :class:`SimulationProblem` instance.
    """

    if base_folder is None:
        # Check command line arguments
        if len(sys.argv) != 2:
            raise Exception("Usage: {} BASE_FOLDER".format(sys.argv[0]))

        base_folder = sys.argv[1]
    else:
        if not os.path.isabs(base_folder):
            # Resolve base folder relative to script folder
            base_folder = os.path.join(sys.path[0], base_folder)

    model_folder = _resolve_folder(kwargs, base_folder, 'model_folder', 'model')
    input_folder = _resolve_folder(kwargs, base_folder, 'input_folder', 'input')
    output_folder = _resolve_folder(kwargs, base_folder, 'output_folder', 'output')

    # Set up logging
    logger = logging.getLogger("rtctools")
    if not logger.hasHandlers() and not any((isinstance(h, logging.StreamHandler) for h in logger.handlers)):
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    # Add pi.DiagHandler, if using PIMixin. Only add it if it does not already exist.
    if (issubclass(simulation_problem_class, SimulationPIMixin) and
            not any((isinstance(h, pi.DiagHandler) for h in logger.handlers))):
        handler = pi.DiagHandler(output_folder)
        logger.addHandler(handler)

    logger.setLevel(log_level)

    logger.info(
        'Using RTC-Tools {}'.format(__version__))
    logger.debug(
        'Using CasADi {}.'.format(casadi.__version__))

    # Run
    prob = simulation_problem_class(
        model_folder=model_folder, input_folder=input_folder, output_folder=output_folder,
        **kwargs)
    prob.simulate()
    return prob



import logging
from typing import Dict, List, Union

import numpy as np

from .optimization_problem import OptimizationProblem

logger = logging.getLogger("rtctools")


class ControlTreeMixin(OptimizationProblem):
    """
    Adds a stochastic control tree to your optimization problem.
    """

    def control_tree_options(self) -> Dict[str, Union[List[str], List[float], int]]:
        """
        Returns a dictionary of options controlling the creation of a k-ary stochastic tree.

        +------------------------+---------------------+-----------------------+
        | Option                 | Type                | Default value         |
        +========================+=====================+=======================+
        | ``forecast_variables`` | ``list`` of strings | All constant inputs   |
        +------------------------+---------------------+-----------------------+
        | ``branching_times``    | ``list`` of floats  | ``self.times()``      |
        +------------------------+---------------------+-----------------------+
        | ``k``                  | ``int``             | ``2``                 |
        +------------------------+---------------------+-----------------------+

        A ``k``-ary tree is generated, branching at every interior branching time.
        Ensemble members are clustered to paths through the tree based on average
        distance over all forecast variables.

        :returns: A dictionary of control tree generation options.
        """

        options = {}

        options['forecast_variables'] = [var.name()
                                         for var in self.dae_variables['constant_inputs']]
        options['branching_times'] = self.times()[1:]
        options['k'] = 2

        return options

    def discretize_controls(self, resolved_bounds):
        # Collect options
        options = self.control_tree_options()

        # Make sure branching times contain initial and final time.  The
        # presence of these is assumed below.
        times = self.times()
        t0 = self.initial_time
        branching_times = options['branching_times']
        n_branching_times = len(branching_times)
        if n_branching_times > len(times) - 1:
            raise Exception("Too many branching points specified")
        branching_times = np.concatenate(([t0], branching_times, [np.inf]))

        logger.debug("ControlTreeMixin: Branching times:")
        logger.debug(branching_times)

        # Branches start at branching times, so that the tree looks like the following:
        #
        #         *-----
        #   *-----
        #         *-----
        #
        #   t0    t1
        #
        # with branching time t1.
        branches = {}

        def branch(current_branch):
            if len(current_branch) >= n_branching_times:
                return

            # Branch stats
            n_branch_members = len(branches[current_branch])
            if n_branch_members == 0:
                # Nothing to do
                return
            distances = np.zeros((n_branch_members, n_branch_members))

            # Decide branching on a segment of the time horizon
            branching_time_0 = branching_times[len(current_branch) + 1]
            branching_time_1 = branching_times[len(current_branch) + 2]

            # Compute reverse ensemble member index-to-distance index map.
            reverse = {}
            for i, member_i in enumerate(branches[current_branch]):
                reverse[member_i] = i

            # Compute distances between ensemble members, summed for all
            # forecast variables
            for forecast_variable in options['forecast_variables']:
                # We assume the time stamps of the forecasts in all ensemble
                # members to be identical
                timeseries = self.constant_inputs(ensemble_member=0)[
                    forecast_variable]
                els = np.logical_and(
                    timeseries.times >= branching_time_0, timeseries.times < branching_time_1)

                # Compute distance between ensemble members
                for i, member_i in enumerate(branches[current_branch]):
                    timeseries_i = self.constant_inputs(ensemble_member=member_i)[
                        forecast_variable]
                    for j, member_j in enumerate(branches[current_branch]):
                        timeseries_j = self.constant_inputs(ensemble_member=member_j)[
                            forecast_variable]
                        distances[
                            i, j] += np.linalg.norm(timeseries_i.values[els] - timeseries_j.values[els])

            # Keep track of ensemble members that have not yet been allocated
            # to a new branch
            available = set(branches[current_branch])

            # We first select the scenario with the max distance to any other branch
            idx = np.argmax(np.amax(distances, axis=0))

            for i in range(options['k']):
                if idx >= 0:
                    branches[current_branch +
                             str(i)] = [branches[current_branch][idx]]

                    available.remove(branches[current_branch][idx])

                    # We select the scenario with the max min distance to the other branches
                    min_distances = np.array([
                        min([np.inf] + [distances[j, k]
                            for j, member_j in enumerate(branches[current_branch])
                            if member_j not in available and member_k in available])
                        for k, member_k in enumerate(branches[current_branch])
                        ], dtype=np.float64)
                    min_distances[np.where(min_distances == np.inf)] = -np.inf

                    idx = np.argmax(min_distances)
                    if min_distances[idx] <= 0:
                        idx = -1
                else:
                    branches[current_branch + str(i)] = []

            # Cluster remaining ensemble members to branches
            for member_i in available:
                min_i = 0
                min_distance = np.inf
                for i in range(options['k']):
                    branch2 = branches[current_branch + str(i)]
                    if len(branch2) > 0:
                        distance = distances[
                            reverse[member_i], reverse[branch2[0]]]
                        if distance < min_distance:
                            min_distance = distance
                            min_i = i
                branches[current_branch + str(min_i)].append(member_i)

            # Recurse
            for i in range(options['k']):
                branch(current_branch + str(i))

        current_branch = ''
        branches[current_branch] = list(range(self.ensemble_size))
        branch(current_branch)

        logger.debug("ControlTreeMixin:  Control tree is:")
        logger.debug(branches)

        # Map ensemble members to control inputs
        # (variable, (ensemble member, step)) -> control_index
        self.__control_indices = [{} for ensemble_member in range(self.ensemble_size)]
        count = 0
        for control_input in self.controls:
            times = self.times(control_input)
            for member in range(self.ensemble_size):
                self.__control_indices[member][control_input] = np.zeros(
                    len(times), dtype=np.int16)
            for branch, members in branches.items():
                if not members:
                    # Avoid making free variables by skipping branches which have no members
                    continue

                branching_time_0 = branching_times[len(branch) + 0]
                branching_time_1 = branching_times[len(branch) + 1]
                els = np.logical_and(
                    times >= branching_time_0, times < branching_time_1)
                nnz = np.count_nonzero(els)
                for member in members:
                    self.__control_indices[member][control_input][els] = \
                        list(range(count, count + nnz))
                count += nnz

        discrete = self._collint_get_discrete(count, self.__control_indices)
        lbx, ubx = self._collint_get_lbx_ubx(count, self.__control_indices)
        x0 = self._collint_get_x0(count, self.__control_indices)

        # Return number of control variables
        return count, discrete, lbx, ubx, x0, self.__control_indices



import logging
import os

import numpy as np

import rtctools.data.csv as csv
from rtctools._internal.caching import cached
from rtctools.simulation.io_mixin import IOMixin

logger = logging.getLogger("rtctools")


class CSVMixin(IOMixin):
    """
    Adds reading and writing of CSV timeseries and parameters to your simulation problem.

    During preprocessing, files named ``timeseries_import.csv``, ``initial_state.csv``,
    and ``parameters.csv`` are read from the ``input`` subfolder.

    During postprocessing, a file named ``timeseries_export.csv`` is written to the ``output`` subfolder.

    :cvar csv_delimiter:           Column delimiter used in CSV files.  Default is ``,``.
    :cvar csv_validate_timeseries: Check consistency of timeseries.  Default is ``True``.
    """

    #: Column delimiter used in CSV files
    csv_delimiter = ','

    #: Check consistency of timeseries
    csv_validate_timeseries = True

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

    def read(self):
        # Call parent class first for default behaviour.
        super().read()

        # Helper function to check if initial state array actually defines
        # only the initial state
        def check_initial_state_array(initial_state):
            """
            Check length of initial state array, throw exception when larger than 1.
            """
            if initial_state.shape:
                raise Exception(
                    'CSVMixin: Initial state file {} contains more than one row of data. '
                    'Please remove the data row(s) that do not describe the initial '
                    'state.'.format(os.path.join(self._input_folder, 'initial_state.csv')))

        # Read CSV files
        _timeseries = csv.load(
            os.path.join(self._input_folder, self.timeseries_import_basename + '.csv'),
            delimiter=self.csv_delimiter, with_time=True)
        self.__timeseries_times = _timeseries[_timeseries.dtype.names[0]]

        self.io.reference_datetime = self.__timeseries_times[0]

        for key in _timeseries.dtype.names[1:]:
            self.io.set_timeseries(
                key,
                self.__timeseries_times,
                np.asarray(_timeseries[key], dtype=np.float64))

        logger.debug("CSVMixin: Read timeseries.")

        try:
            _parameters = csv.load(
                os.path.join(self._input_folder, 'parameters.csv'),
                delimiter=self.csv_delimiter)
            for key in _parameters.dtype.names:
                self.io.set_parameter(key, float(_parameters[key]))
            logger.debug("CSVMixin: Read parameters.")
        except IOError:
            pass

        try:
            _initial_state = csv.load(
                os.path.join(self._input_folder, 'initial_state.csv'),
                delimiter=self.csv_delimiter)
            logger.debug("CSVMixin: Read initial state.")
            check_initial_state_array(_initial_state)
            self.__initial_state = {
                key: float(_initial_state[key]) for key in _initial_state.dtype.names}
        except IOError:
            self.__initial_state = {}

        # Check for collisions in __initial_state and timeseries import (CSV)
        for collision in set(self.__initial_state) & set(_timeseries.dtype.names[1:]):
            if self.__initial_state[collision] == _timeseries[collision][0]:
                continue
            else:
                logger.warning(
                    'CSVMixin: Entry {} in initial_state.csv conflicts with '
                    'timeseries_import.csv'.format(collision))

        # Timestamp check
        if self.csv_validate_timeseries:
            times = self.__timeseries_times
            for i in range(len(times) - 1):
                if times[i] >= times[i + 1]:
                    raise Exception(
                        'CSVMixin: Time stamps must be strictly increasing.')

        times = self.__timeseries_times
        dt = times[1] - times[0]

        # Check if the timeseries are truly equidistant
        if self.csv_validate_timeseries:
            for i in range(len(times) - 1):
                if times[i + 1] - times[i] != dt:
                    raise Exception(
                        'CSVMixin: Expecting equidistant timeseries, the time step '
                        'towards {} is not the same as the time step(s) before. '
                        'Set equidistant=False if this is intended.'.format(
                            times[i + 1]))

    def write(self):
        # Call parent class first for default behaviour.
        super().write()

        times = self._simulation_times

        # Write output
        names = ['time'] + sorted(set(self._io_output_variables))
        formats = ['O'] + (len(names) - 1) * ['f8']
        dtype = {'names': names, 'formats': formats}
        data = np.zeros(len(times), dtype=dtype)
        data['time'] = self.io.sec_to_datetime(times, self.io.reference_datetime)
        for variable in self._io_output_variables:
            data[variable] = np.array(self._io_output[variable])

        fname = os.path.join(self._output_folder, self.timeseries_export_basename + '.csv')
        csv.save(fname, data, delimiter=self.csv_delimiter, with_time=True)

    @cached
    def initial_state(self):
        """
        The initial state. Includes entries from parent classes and initial_state.csv

        :returns: A dictionary of variable names and initial state (t0) values.
        """
        # Call parent class first for default values.
        initial_state = super().initial_state()

        # Set of model vars that are allowed to have an initial state
        valid_model_vars = set(self.get_state_variables()) | set(self.get_input_variables())

        # Load initial states from __initial_state
        for variable, value in self.__initial_state.items():

            # Get the cannonical vars and signs
            canonical_var, sign = self.alias_relation.canonical_signed(variable)

            # Only store variables that are allowed to have an initial state
            if canonical_var in valid_model_vars:
                initial_state[canonical_var] = value * sign

                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("CSVMixin: Read initial state {} = {}".format(variable, value))
            else:
                logger.warning("CSVMixin: In initial_state.csv, {} is not an input or state variable.".format(variable))
        return initial_state



from casadi import if_else, logic_and

from .bspline import BSpline


class BSpline2D(BSpline):
    """
    Arbitrary order, two-dimensional, non-uniform B-Spline.
    """

    def __init__(self, tx, ty, w, kx=3, ky=3):
        """
        Create a new 2D B-Spline object.

        :param tx: Knot vector in X direction.
        :param ty: Knot vector in Y direction.
        :param w:  Weight vector.
        :param kx: Spline order in X direction.
        :param ky: Spline order in Y direction.
        """

        # Store arguments
        self.__tx = tx
        self.__ty = ty
        self.__w = w
        self.__kx = kx
        self.__ky = ky

    def __call__(self, x, y):
        """
        Evaluate the B-Spline at point (x, y).

        The support of this function is the half-open interval [tx[0], tx[-1]) x [ty[0], ty[-1]).

        :param x: The coordinate of the point at which to evaluate.
        :param y: The ordinate of the point at which to evaluate.

        :returns: The spline evaluated at the given point.
        """
        z = 0.0
        for i in range(len(self.__tx) - self.__kx - 1):
            bx = if_else(logic_and(x >= self.__tx[i], x <= self.__tx[
                         i + self.__kx + 1]), self.basis(self.__tx, x, self.__kx, i), 0.0)
            for j in range(len(self.__ty) - self.__ky - 1):
                by = if_else(logic_and(y >= self.__ty[j], y <= self.__ty[
                             j + self.__ky + 1]), self.basis(self.__ty, y, self.__ky, j), 0.0)
                z += self.__w[i * (len(self.__ty) - self.__ky - 1) + j] * bx * by
        return z



import logging
import os
from datetime import timedelta

import numpy as np

import rtctools.data.csv as csv
from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools.optimization.io_mixin import IOMixin
from rtctools.optimization.timeseries import Timeseries

logger = logging.getLogger("rtctools")


class CSVMixin(IOMixin):
    """
    Adds reading and writing of CSV timeseries and parameters to your optimization problem.

    During preprocessing, files named ``timeseries_import.csv``, ``initial_state.csv``,
    and ``parameters.csv`` are read from the ``input`` subfolder.

    During postprocessing, a file named ``timeseries_export.csv`` is written to the ``output`` subfolder.

    In ensemble mode, a file named ``ensemble.csv`` is read from the ``input`` folder.  This file
    contains two columns. The first column gives the name of the ensemble member, and the second
    column its probability.  Furthermore, the other XML files appear one level deeper inside the
    filesystem hierarchy, inside subfolders with the names of the ensemble members.

    :cvar csv_initial_state_basename:
        Initial state file basename. Default is ``initial_state``.
    :cvar csv_parameters_basename:
        Parameters file basename. Default is ``parameters``.
    :cvar csv_ensemble_basename:
        Ensemble file basename. Default is ``ensemble``.
    :cvar csv_delimiter:
        Column delimiter used in CSV files.  Default is ``,``.
    :cvar csv_equidistant:
        Whether or not the timeseries data is equidistant.  Default is ``True``.
    :cvar csv_ensemble_mode:
        Whether or not to use ensembles.  Default is ``False``.
    :cvar csv_validate_timeseries:
        Check consistency of timeseries.  Default is ``True``.
    """

    #: Initial state file basename
    csv_initial_state_basename = 'initial_state'

    #: Parameters file basename
    csv_parameters_basename = 'parameters'

    #: Ensemble file basename
    csv_ensemble_basename = 'ensemble'

    #: Column delimiter used in CSV files
    csv_delimiter = ','

    #: Whether or not the timeseries data is equidistant
    csv_equidistant = True

    #: Whether or not to use ensembles
    csv_ensemble_mode = False

    #: Check consistency of timeseries
    csv_validate_timeseries = True

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

    def read(self):
        # Call parent class first for default behaviour.
        super().read()

        # Helper function to check if initial state array actually defines
        # only the initial state
        def check_initial_state_array(initial_state):
            """
            Check length of initial state array, throw exception when larger than 1.
            """
            if initial_state.shape:
                raise Exception(
                    'CSVMixin: Initial state file {} contains more than one row of data. '
                    'Please remove the data row(s) that do not describe the initial state.'.format(
                        os.path.join(self._input_folder, self.csv_initial_state_basename + '.csv')))

        # Read CSV files
        self.__initial_state = []
        if self.csv_ensemble_mode:
            self.__ensemble = np.genfromtxt(
                os.path.join(self._input_folder, self.csv_ensemble_basename + '.csv'),
                delimiter=",", deletechars='', dtype=None, names=True, encoding=None)

            logger.debug("CSVMixin: Read ensemble description")

            for ensemble_member_index, ensemble_member_name in enumerate(self.__ensemble['name']):
                _timeseries = csv.load(
                    os.path.join(
                        self._input_folder,
                        ensemble_member_name,
                        self.timeseries_import_basename + ".csv",
                    ),
                    delimiter=self.csv_delimiter,
                    with_time=True,
                )
                self.__timeseries_times = _timeseries[_timeseries.dtype.names[0]]

                self.io.reference_datetime = self.__timeseries_times[0]

                for key in _timeseries.dtype.names[1:]:
                    self.io.set_timeseries(
                        key,
                        self.__timeseries_times,
                        np.asarray(_timeseries[key], dtype=np.float64),
                        ensemble_member_index
                    )
            logger.debug("CSVMixin: Read timeseries")

            for ensemble_member_index, ensemble_member_name in enumerate(self.__ensemble['name']):
                try:
                    _parameters = csv.load(os.path.join(
                        self._input_folder, ensemble_member_name, self.csv_parameters_basename + '.csv'),
                        delimiter=self.csv_delimiter)
                    for key in _parameters.dtype.names:
                        self.io.set_parameter(key, float(_parameters[key]), ensemble_member_index)
                except IOError:
                    pass
            logger.debug("CSVMixin: Read parameters.")

            for ensemble_member_name in self.__ensemble['name']:
                try:
                    _initial_state = csv.load(os.path.join(
                        self._input_folder, ensemble_member_name, self.csv_initial_state_basename + '.csv'),
                        delimiter=self.csv_delimiter)
                    check_initial_state_array(_initial_state)
                    _initial_state = {key: float(_initial_state[key]) for key in _initial_state.dtype.names}
                except IOError:
                    _initial_state = {}
                self.__initial_state.append(AliasDict(self.alias_relation, _initial_state))
            logger.debug("CSVMixin: Read initial state.")
        else:
            _timeseries = csv.load(
                os.path.join(
                    self._input_folder, self.timeseries_import_basename + ".csv"
                ),
                delimiter=self.csv_delimiter,
                with_time=True,
            )
            self.__timeseries_times = _timeseries[_timeseries.dtype.names[0]]

            self.io.reference_datetime = self.__timeseries_times[0]

            for key in _timeseries.dtype.names[1:]:
                self.io.set_timeseries(key, self.__timeseries_times, np.asarray(_timeseries[key], dtype=np.float64))
            logger.debug("CSVMixin: Read timeseries.")

            try:
                _parameters = csv.load(os.path.join(
                    self._input_folder, self.csv_parameters_basename + '.csv'), delimiter=self.csv_delimiter)
                logger.debug("CSVMixin: Read parameters.")
                for key in _parameters.dtype.names:
                    self.io.set_parameter(key, float(_parameters[key]))
            except IOError:
                pass

            try:
                _initial_state = csv.load(os.path.join(
                    self._input_folder, self.csv_initial_state_basename + '.csv'), delimiter=self.csv_delimiter)
                logger.debug("CSVMixin: Read initial state.")
                check_initial_state_array(_initial_state)
                _initial_state = {key: float(_initial_state[key]) for key in _initial_state.dtype.names}
            except IOError:
                _initial_state = {}
            self.__initial_state.append(AliasDict(self.alias_relation, _initial_state))

        # Timestamp check
        if self.csv_validate_timeseries:
            times = self.__timeseries_times
            for i in range(len(times) - 1):
                if times[i] >= times[i + 1]:
                    raise Exception(
                        'CSVMixin: Time stamps must be strictly increasing.')

        if self.csv_equidistant:
            # Check if the timeseries are truly equidistant
            if self.csv_validate_timeseries:
                times = self.__timeseries_times
                dt = times[1] - times[0]
                for i in range(len(times) - 1):
                    if times[i + 1] - times[i] != dt:
                        raise Exception(
                            'CSVMixin: Expecting equidistant timeseries, the time step towards '
                            '{} is not the same as the time step(s) before. Set csv_equidistant = False '
                            'if this is intended.'.format(times[i + 1]))

    def ensemble_member_probability(self, ensemble_member):
        if self.csv_ensemble_mode:
            return self.__ensemble['probability'][ensemble_member]
        else:
            return 1.0

    @cached
    def history(self, ensemble_member):
        # Call parent class first for default values.
        history = super().history(ensemble_member)

        initial_time = np.array([self.initial_time])

        # Load parameters from parameter config
        for variable in self.dae_variables['free_variables']:
            variable = variable.name()
            try:
                history[variable] = Timeseries(initial_time,
                                               self.__initial_state[ensemble_member][variable])
            except (KeyError, ValueError):
                pass
            else:
                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("CSVMixin: Read initial state {}".format(variable))
        return history

    def write(self):
        # Call parent class first for default behaviour.
        super().write()

        # Write output
        times = self.times()

        def write_output(ensemble_member, folder):
            results = self.extract_results(ensemble_member)
            names = ['time'] + sorted({sym.name() for sym in self.output_variables})
            formats = ['O'] + (len(names) - 1) * ['f8']
            dtype = {'names': names, 'formats': formats}
            data = np.zeros(len(times), dtype=dtype)
            data['time'] = [self.io.reference_datetime + timedelta(seconds=s) for s in times]
            for output_variable in self.output_variables:
                output_variable = output_variable.name()
                try:
                    values = results[output_variable]
                    if len(values) != len(times):
                        values = self.interpolate(
                            times, self.times(output_variable), values, self.interpolation_method(output_variable))
                except KeyError:
                    try:
                        ts = self.get_timeseries(
                            output_variable, ensemble_member)
                        if len(ts.times) != len(times):
                            values = self.interpolate(
                                times, ts.times, ts.values)
                        else:
                            values = ts.values
                    except KeyError:
                        logger.error(
                            "Output requested for non-existent variable {}".format(output_variable))
                        continue
                data[output_variable] = values

            fname = os.path.join(folder, self.timeseries_export_basename + ".csv")
            csv.save(fname, data, delimiter=self.csv_delimiter, with_time=True)

        if self.csv_ensemble_mode:
            for ensemble_member, ensemble_member_name in enumerate(self.__ensemble['name']):
                write_output(ensemble_member, os.path.join(
                    self._output_folder, ensemble_member_name))
        else:
            write_output(0, self._output_folder)



import logging
import os
import shutil
import sys
from pathlib import Path

import rtctools

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')
logger = logging.getLogger("rtctools")
logger.setLevel(logging.INFO)


def copy_libraries(*args):

    if not args:
        args = sys.argv[1:]

    if not args:
        path = input("Folder to put the Modelica libraries: [.] ") or "."
    else:
        path = args[0]

    if not os.path.exists(path):
        sys.exit("Folder '{}' does not exist".format(path))

    # pkg_resources can be quite a slow import, so we do it here
    import pkg_resources

    def _copytree(src, dst, symlinks=False, ignore=None):
        if not os.path.exists(dst):
            os.makedirs(dst)
        for item in os.listdir(src):
            s = os.path.join(src, item)
            d = os.path.join(dst, item)
            if os.path.isdir(s):
                _copytree(s, d, symlinks, ignore)
            else:
                if not os.path.exists(d):
                    shutil.copy2(s, d)
                elif Path(s).name.lower() == 'package.mo':
                    # Pick the largest one, assuming that all plugin packages
                    # to not provide a meaningful package.mo
                    if os.stat(s).st_size > os.stat(d).st_size:
                        logger.warning("Overwriting '{}' with '{}' as the latter is larger.".format(d, s))
                        os.remove(d)
                        shutil.copy2(s, d)
                    else:
                        logger.warning("Not copying '{}' to '{}' as the latter is larger.".format(s, d))
                else:
                    raise OSError("Could not combine two folders")

    dst = Path(path)

    library_folders = []
    for ep in pkg_resources.iter_entry_points(group='rtctools.libraries.modelica'):
        if ep.name == "library_folder":
            library_folders.append(
                Path(pkg_resources.resource_filename(ep.module_name, ep.attrs[0])))

    tlds = {}
    for lf in library_folders:
        for x in lf.iterdir():
            if x.is_dir():
                tlds.setdefault(x.name, []).append(x)

    for tld, paths in tlds.items():
        if Path(tld).exists():
            sys.exit("Library with name '{}'' already exists".format(tld))

        try:
            for p in paths:
                _copytree(p, dst / p.name)
        except OSError:
            sys.exit("Failed merging the libraries in package '{}'".format(tld))

    sys.exit("Succesfully copied all library folders to '{}'".format(dst.resolve()))


def download_examples(*args):

    if not args:
        args = sys.argv[1:]

    if not args:
        path = input("Folder to download the examples to: [.] ") or "."
    else:
        path = args[0]

    if not os.path.exists(path):
        sys.exit("Folder '{}' does not exist".format(path))

    path = Path(path)

    import urllib.request
    from urllib.error import HTTPError
    from zipfile import ZipFile

    # GitLab is blocking requests unless we specify a user agent
    user_agent = 'Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:47.0) Gecko/20100101 Firefox/47.0'

    version = rtctools.__version__
    rtc_full_name = 'rtc-tools-{}'.format(version)
    try:
        url = 'https://gitlab.com/deltares/rtc-tools/-/archive/' \
            '{}/{}.zip'.format(version, rtc_full_name)

        opener = urllib.request.build_opener()
        opener.addheaders = [('User-agent', user_agent)]
        urllib.request.install_opener(opener)
        local_filename, _ = urllib.request.urlretrieve(url)
    except HTTPError:
        sys.exit("Could not found examples for RTC-Tools version {}.".format(version))

    with ZipFile(local_filename, 'r') as z:
        target = path / 'rtc-tools-examples'
        prefix = '{}/examples/'.format(rtc_full_name)
        members = [x for x in z.namelist() if x.startswith(prefix)]
        z.extractall(members=members)
        shutil.move(prefix, target)
        shutil.rmtree(rtc_full_name)

        sys.exit("Succesfully downloaded the RTC-Tools examples to '{}'".format(target.resolve()))

    try:
        os.remove(local_filename)
    except OSError:
        pass



from casadi import Function, SX, if_else, inf, jacobian, logic_and, nlpsol, sum2, vertcat

import numpy as np

from .bspline import BSpline


class BSpline1D(BSpline):
    """
    Arbitrary order, one-dimensional, non-uniform B-Spline implementation using Cox-de Boor recursion.
    """

    def __init__(self, t, w, k=3):
        """
        Create a new 1D B-Spline object.

        :param t: Knot vector.
        :param w: Weight vector.
        :param k: Spline order.
        """

        # Store arguments
        self.__t = t
        self.__w = w
        self.__k = k

    def __call__(self, x):
        """
        Evaluate the B-Spline at point x.

        The support of this function is the half-open interval [t[0], t[-1]).

        :param x: The point at which to evaluate.

        :returns: The spline evaluated at the given point.
        """
        y = 0.0
        for i in range(len(self.__t) - self.__k - 1):
            y += if_else(logic_and(x >= self.__t[i], x <= self.__t[i + self.__k + 1]), self.__w[
                         i] * self.basis(self.__t, x, self.__k, i), 0.0)
        return y

    @classmethod
    def fit(cls, x, y, k=3, monotonicity=0, curvature=0,
            num_test_points=100, epsilon=1e-7, delta=1e-4, interior_pts=None):
        """
        fit() returns a tck tuple like scipy.interpolate.splrep, but adjusts
        the weights to meet the desired constraints to the curvature of the spline curve.

        :param monotonicity:
            - is an integer, magnitude is ignored
            - if positive, causes spline to be monotonically increasing
            - if negative, causes spline to be monotonically decreasing
            - if 0, leaves spline monotonicity unconstrained

        :param curvature:
            - is an integer, magnitude is ignored
            - if positive, causes spline curvature to be positive (convex)
            - if negative, causes spline curvature to be negative (concave)
            - if 0, leaves spline curvature unconstrained

        :param num_test_points:
            - sets the number of points that the constraints will be applied at across
              the range of the spline

        :param epsilon:
            - offset of monotonicity and curvature constraints from zero, ensuring strict
              monotonicity
            - if epsilon is set to less than the tolerance of the solver, errors will result

        :param delta:
            - amount the first and last knots are extended outside the range of the splined points
            - ensures that the spline evaluates correctly at the first and last nodes, as
              well as the distance delta beyond these nodes

        :param interior_pts:
            - optional list of interior knots to use

        :returns: A tuple of spline knots, weights, and order.
        """
        x = np.asarray(x)
        y = np.asarray(y)
        N = len(x)

        if interior_pts is None:
            # Generate knots: This algorithm is based on the Fitpack algorithm by p.dierckx
            # The original code lives here: http://www.netlib.org/dierckx/
            if k % 2 == 1:
                interior_pts = x[k // 2 + 1:-k // 2]
            else:
                interior_pts = (x[k // 2 + 1:-k // 2] + x[k // 2:-k // 2 - 1]) / 2
        t = np.concatenate(
            (np.full(k + 1, x[0] - delta), interior_pts, np.full(k + 1, x[-1] + delta)))
        num_knots = len(t)

        # Casadi Variable Symbols
        c = SX.sym('c', num_knots)
        x_sym = SX.sym('x')

        # Casadi Representation of Spline Function & Derivatives
        expr = cls(t, c, k)(x_sym)
        free_vars = [c, x_sym]
        bspline = Function('bspline', free_vars, [expr])
        J = jacobian(expr, x_sym)
        # bspline_prime = Function('bspline_prime', free_vars, [J])
        H = jacobian(J, x_sym)
        bspline_prime_prime = Function('bspline_prime_prime', free_vars, [H])

        # Objective Function
        xpt = SX.sym('xpt')
        ypt = SX.sym('ypt')
        sq_diff = Function('sq_diff', [xpt, ypt], [
                             (ypt - bspline(c, xpt))**2])
        sq_diff = sq_diff.map(N, 'serial')
        f = sum2(sq_diff(SX(x), SX(y)))

        # Setup Curvature Constraints
        delta_c_max = np.full(num_knots - 1, inf)
        delta_c_min = np.full(num_knots - 1, -inf)
        max_slope_slope = np.full(num_test_points, inf)
        min_slope_slope = np.full(num_test_points, -inf)
        if monotonicity != 0:
            if monotonicity < 0:
                delta_c_max = np.full(num_knots - 1, -epsilon)
            else:
                delta_c_min = np.full(num_knots - 1, epsilon)
        if curvature != 0:
            if curvature < 0:
                max_slope_slope = np.full(num_test_points, -epsilon)
            else:
                min_slope_slope = np.full(num_test_points, epsilon)
        monotonicity_constraints = vertcat(*[
            c[i + 1] - c[i] for i in range(num_knots - 1)])
        x_linspace = np.linspace(x[0], x[-1], num_test_points)
        curvature_constraints = vertcat(*[
            bspline_prime_prime(c, SX(x)) for x in x_linspace])
        g = vertcat(monotonicity_constraints, curvature_constraints)
        lbg = np.concatenate((delta_c_min, min_slope_slope))
        ubg = np.concatenate((delta_c_max, max_slope_slope))

        # Perform mini-optimization problem to calculate the the values of c
        nlp = {'x': c, 'f': f, 'g': g}
        my_solver = "ipopt"
        solver = nlpsol("solver", my_solver, nlp, {'print_time': 0, 'expand': True, 'ipopt': {'print_level': 0}})
        sol = solver(lbg=lbg, ubg=ubg)
        stats = solver.stats()
        return_status = stats['return_status']
        if return_status not in ['Solve_Succeeded', 'Solved_To_Acceptable_Level', 'SUCCESS']:
            raise Exception("Spline fitting failed with status {}".format(return_status))

        # Return the new tck tuple
        return (t, np.array(sol['x']).ravel(), k)
